/***************************************************************************
                          dlgborder.cpp  -  description
                             -------------------
    begin                : Thu Apr 26 2001
    copyright            : (C) 2001 by benny
    email                : bm@cage.rug.ac.be
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

//Qt includes
#include <qpainter.h>
#include <qlineedit.h>
#include <qvalidator.h>
#include <qradiobutton.h>
#include <qcheckbox.h>
#include <qpushbutton.h>
#include <qlcdnumber.h>
#include <qmessagebox.h>

//KDE includes
#include <klocale.h>

//application includes
#include "dlgborder.h"
#include "k2ddomain.h"
  //always call als modal
dlgBorder::dlgBorder(K2dDomain * dom, QWidget *parent, const char* name) : dlgBorderDesign(parent, name , true, 0)
{
    //set the domain
    dlgDomain = dom;  // refer to same data !
  /** //for points only doubles are allowed, not more than 3 after , */
    ledP1x->setValidator(new QDoubleValidator(-999999.0, 999999.0, 3, ledP1x));
    ledP1z->setValidator(new QDoubleValidator(-999999.0, 999999.0, 3, ledP1z));
    ledP2x->setValidator(new QDoubleValidator(-999999.0, 999999.0, 3, ledP2x));
    ledP2z->setValidator(new QDoubleValidator(-999999.0, 999999.0, 3, ledP2z));
    // parameters sinus curve
    led_L->setValidator(new QDoubleValidator(-999999.0, 999999.0, 3, led_L));
    led_a->setValidator(new QDoubleValidator(-99999.0, 99999.0, 3, led_a));
    led_b->setValidator(new QDoubleValidator(-99999.0, 99999.0, 3, led_b));
    led_V->setValidator(new QDoubleValidator(-999.0, 999.0, 3, led_V));
    led_n->setValidator(new QIntValidator(0, 99, led_n));
    ledSinMeshInt->setValidator(new QIntValidator(0, 9999, ledSinMeshInt));
  /** //for triangles only doubles are allowed, positive, not more than 3 after , and not extremely large */
    ledBeginTriaSize->setValidator(new QDoubleValidator(0.0009,  999.9,  3,  ledBeginTriaSize));
    ledEndTriaSize->setValidator(new QDoubleValidator(0.0009,  999.9,  3,  ledEndTriaSize));
    ledSinTriangle->setValidator(new QDoubleValidator(0.0009,  999.9,  3,  ledSinTriangle));
  /** //for parameters a and b only doubles are allowed, not more than 3 after ,  not extreme*/
    ledA->setValidator(new QDoubleValidator(-10000.0, 10000.0, 3, ledA));
    ledB->setValidator(new QDoubleValidator(-10000.0, 10000.0, 3, ledB));

   //do connections between new slots and buttons
   connect( cmdSave, SIGNAL(clicked()), this, SLOT(slotBorderSave()) );
   connect( cmdSaveExit, SIGNAL(clicked()), this, SLOT(slotSaveExit()) );
   connect( cmdPrevious, SIGNAL(clicked()), this, SLOT(slotPreviousBorder()) );
   connect( cmdNext, SIGNAL(clicked()), this, SLOT(slotNextBorder()) );
   connect( cmdAddEnd, SIGNAL(clicked()), this, SLOT(slotAddEnd()) );
   connect( cmdDelete, SIGNAL(clicked()), this, SLOT(slotDelete()) );

   //finish with loading the first border in the view
   loadBorder(1);

}
dlgBorder::~dlgBorder(){
}
/** Validation of given values in the dialog */
QString dlgBorder::validateInput(){
  //is a bordertype selected ?
  if (! ( rbtLine->isChecked() || rbtSinCurve->isChecked() || rbtCurve->isChecked() ) )
    {return i18n("Select a  Border Type please. ");}
  //is a border condition selected
  if (! ( rbtNormal->isChecked() || rbtDirichlet->isChecked() ) )
    {return i18n("Select a  Border Condition please. ");}
  //validate input if line is selected
  int i=1 ;
  QString text;
  if (rbtLine->isChecked() )
   {
    text =  ledP1x->text() ;
    if ( !(ledP1x->validator()->validate( text , i)  == QValidator::Acceptable ) )
      {return i18n("Point Begin x not Acceptable! \n"
                              "  Take value between -999999.000 and 999999.000 \n "
                              "   Maximum 3 decimals. ");
      }
    text =  ledP1z->text() ;
    if ( !(ledP1z->validator()->validate( text , i)  == QValidator::Acceptable ) )
      {return i18n("Point Begin z not Acceptable! \n"
                              "  Take value between -999999.000 and 999999.000 \n "
                              "   Maximum 3 decimals. ");
      }
    text =  ledP2x->text() ;
    if ( !(ledP2x->validator()->validate( text , i)  == QValidator::Acceptable ) )
      {return i18n("Point End x not Acceptable! \n"
                              "  Take value between -999999.000 and 999999.000 \n "
                              "   Maximum 3 decimals. ");
      }
    text =  ledP2z->text() ;
    if ( !(ledP2z->validator()->validate( text , i)  == QValidator::Acceptable ) )
      {return i18n("Point End z not Acceptable! \n"
                              "  Take value between -999999.000 and 999999.000 \n "
                              "   Maximum 3 decimals. ");
      }
   }
  //validate input if sinus curve selected
  if (rbtSinCurve->isChecked() )
   {
    text =  led_L->text() ;
    if ( !(led_L->validator()->validate( text , i)  == QValidator::Acceptable ) )
      {return i18n("Length L not Acceptable! \n"
                              "  Take value between -999999.000 and 999999.000 \n "
                              "   Maximum 3 decimals. ");
      }
    // L may not be 0 or too small
    if (text.toDouble() < 0.1 && text.toDouble() > -0.1)
      {return i18n("Length L to small \n"
                              "  Take value not out of range -0.1 and 0.1!\n "
                              "   Maximum 3 decimals. ");
      }
    text =  led_a->text() ;
    if ( !(led_a->validator()->validate( text , i)  == QValidator::Acceptable ) )
      {return i18n("Slope a not Acceptable! \n"
                              "  Take value between -99999.000 and 99999.000 \n "
                              "   Maximum 3 decimals. ");
      }
    text =  led_b->text() ;
    if ( !(led_b->validator()->validate( text , i)  == QValidator::Acceptable ) )
      {return i18n("Offset b not Acceptable! \n"
                              "  Take value between -99999.000 and 99999.000 \n "
                              "   Maximum 3 decimals. ");
      }
    text =  led_V->text() ;
    if ( !(led_V->validator()->validate( text , i)  == QValidator::Acceptable ) )
      {return i18n("Amplitude V not Acceptable! \n"
                              "  Take value between -999.000 and 999.000 \n "
                              "   Maximum 3 decimals. ");
      }
    text =  led_n->text() ;
    if ( !(led_n->validator()->validate( text , i)  == QValidator::Acceptable ) )
      {return i18n("Number of Bumps (Hills) n not Acceptable! \n"
                              "  Take positive integer between 0 and 99 \n ");
      }
    text =  ledSinMeshInt->text() ;
    if ( !(ledSinMeshInt->validator()->validate( text , i)  == QValidator::Acceptable ) )
      {return i18n("Number of mesh intervals in sine curveBumps is not Acceptable! \n"
                              "  Take positive integer between 0 and 9999 \n ");
      }
    text =  ledSinTriangle->text() ;
    if ( !(ledSinTriangle->validator()->validate( text , i)  == QValidator::Acceptable ) )
      {return i18n("Triangle size on sinus curve not Acceptable! \n"
                              "  Take value between 0.0009,  999.9 . \n "
                              "   Maximum 3 decimals. ");
      }
   }
  //validate if line is checked that begin and end point are nog equal
  if (rbtLine->isChecked() )
   {
    if (( ledP1x->text().toDouble() == ledP2x->text().toDouble()) && (ledP1z->text().toDouble() == ledP2z->text().toDouble() ) )
      {return i18n("Begin Point is Equal to End point ! \n"
                              "  Take different values. " );
      }
   }
  //validate triangle sides if line is selected
  if (rbtLine->isChecked() )
   {
    int i=1 ;
    QString text;
    text =  ledBeginTriaSize->text() ;
    if ( !(ledBeginTriaSize->validator()->validate( text , i)  == QValidator::Acceptable ) )
      {return i18n("Triangle size point begin not Acceptable! \n"
                              "  Take value between 0.0009,  999.9 . \n "
                              "   Maximum 3 decimals. ");
      }
    text =  ledEndTriaSize->text() ;
    if ( !(ledEndTriaSize->validator()->validate( text , i)  == QValidator::Acceptable ) )
      {return i18n("Triangle size point end not Acceptable! \n"
                              "  Take value between 0.0009,  999.9 . \n "
                              "   Maximum 3 decimals. ");
      }
   }
  //validate if function is given when dirichletcondition is chosen
  if (rbtDirichlet->isChecked() )
   {
     if (! ( rbtZ->isChecked() || rbtAx_b->isChecked() ) )
       {return i18n("Select a  Function for the DirichletCondition please. ");}
   }
  //validate if a and b parameter are well typed in if dirichletcondition is chosen and function ax+b
  if (rbtDirichlet->isChecked() && rbtAx_b->isChecked() )
   {
    int i=1 ;
    QString text;
    text =  ledA->text() ;
    if ( !(ledA->validator()->validate( text , i)  == QValidator::Acceptable ) )
      {return i18n("The given a parameter is not Acceptable! \n"
                              "  Take value between -10000,  10000 . \n "
                              "   Maximum 3 decimals. ");
      }
    text =  ledB->text() ;
    if ( !(ledB->validator()->validate( text , i)  == QValidator::Acceptable ) )
      {return i18n("The given b parameter is not Acceptable! \n"
                              "  Take value between -10000,  10000 . \n "
                              "   Maximum 3 decimals. ");
      }
   }
  //everything ok
  return "";
}

void dlgBorder::loadBorder(uint index)
{
  //set lcdnumber to the border we are loading
  int intIndex = index;
  lcdNr->display(intIndex);
  //start with init of everything on screen
  //init bordertype values
  ledP1x->setText("");
  ledP1z->setText("");
  ledP2x->setText("");
  ledP2z->setText("");
  ledBeginTriaSize->setText("1");
  ledEndTriaSize->setText("1");
  led_L->setText("");
  led_a->setText("");
  led_b->setText("");
  led_V->setText("");
  led_n->setText("");
  ledSinMeshInt->setText("0");
  ledSinTriangle->setText("1");
  chbReverse->setChecked(false);
  //init bordercond values, we put standard the condition on normal derivative 0
  ledA->setText("");
  ledB->setText("");
  //now load what's present
  if (index > dlgDomain->nrOfBorders())
     { //new border, put condition to normal, so user does not have to do it explicitely
        rbtNormal->setChecked(true);
     } else {
       //make an interator to the border we need in the domain
       //first make an iterator that points to the beginning (index=0)
      QListIterator<K2dBorderBase> iterDom =  dlgDomain->iterator();
      //make iterator point at index - 1 position
      for (uint i=0 ; i < (index - 1)  ; ++i )
          { ++iterDom ;
          }
       if (iterDom.current()->type() == K2dBorderBase::Line )
          { //we have a line as input, so set all the stuff for that
            QString iStr;
            //since we now it's a line, we can do a dynamic cast to access line specific data
            K2dBorderLine* tempHelp = dynamic_cast <K2dBorderLine *> ( iterDom.current() );
            rbtLine->setChecked(true);
            ledP1x->setText(iStr.setNum( iterDom.current()->firstPoint()->at(0) ));
            ledP1z->setText(iStr.setNum( iterDom.current()->firstPoint()->at(1) ));
            ledP2x->setText(iStr.setNum( iterDom.current()->lastPoint()->at(0) ));
            ledP2z->setText(iStr.setNum( iterDom.current()->lastPoint()->at(1) ));
            ledBeginTriaSize->setText(iStr.setNum( tempHelp->triaSize(K2dBorderLine::BeginPoint) ));
            ledEndTriaSize->setText(iStr.setNum( tempHelp->triaSize(K2dBorderLine::EndPoint) ));
          }
       if (iterDom.current()->type() == K2dBorderBase::SinCurve )
          { //we have a sinecurve as input, so set all the stuff for that
            QString iStr;
            //since we now it's a sinecurve, we can do a dynamic cast to access sine specific data
            K2dBorderSinCurve* tempHelp = dynamic_cast <K2dBorderSinCurve *> ( iterDom.current() );
            rbtSinCurve->setChecked(true);
            led_L->setText(iStr.setNum( tempHelp->getLength() ));
            led_V->setText(iStr.setNum( tempHelp->getAmplitude() ));
            led_a->setText(iStr.setNum( tempHelp->getSlope() ));
            led_b->setText(iStr.setNum( tempHelp->getOffset() ));
            led_n->setText(iStr.setNum( tempHelp->getHills() ));
            ledSinMeshInt->setText(iStr.setNum( tempHelp->nrMeshIntervalPoints() ));
            ledSinTriangle->setText(iStr.setNum( tempHelp->triaSize() ));
            chbReverse->setChecked(tempHelp->reverseValue());
          }
       if (iterDom.current()->getBorderCondition() == K2dBorderBase::NEUMANN )
          { //we have  the condition that the normal derivative is 0 (that's the only neumann cond implemented), so set all the stuff for that
            QString iStr;
            rbtNormal->setChecked(true);
          } else {
            if (iterDom.current()->getBorderCondition() == K2dBorderBase::DIRICHLET )
              {
                QString iStr;
                rbtDirichlet->setChecked(true);
                //check the type of dirichlet condition. For now only 2 possible
                if (iterDom.current()->getBorderCondDirType() == K2dBorderBase::DirZ )
                  {  rbtZ->setChecked(true);
                  } else {
                     rbtAx_b->setChecked(true);
                     ledA->setText(iStr.setNum( iterDom.current()->getBorderCondDirLineA() ));
                     ledB->setText(iStr.setNum( iterDom.current()->getBorderCondDirLineB() ));
                  }
              }
          }
     }
}

/** Save border cklicked, ok is set to false if saving is unsuccessfulle */
bool dlgBorder::slotBorderSave( ) {
  //first validate if input is ok. If so, do actual saving
  QString validationString = validateInput();

  if (!validationString.isEmpty())
    {QMessageBox::information( this, "K2dGrFlow",
                            validationString );
      //stop saving process
      return false;
    } ;
  //everything is ok, we save the border in the domain, and generate a save to disk in the application
  //first we delete the old border, then we add the new
  uint index= lcdNr->intValue();
  if ( index <= dlgDomain->nrOfBorders() )
     { dlgDomain->removeBorder(index-1);
     }
  //see type of border we want to save, and do the necessary for that
  if (rbtLine->isChecked())
    {
     BorderPoint a(ledP1x->text().toDouble(), ledP1z->text().toDouble());
     BorderPoint b(ledP2x->text().toDouble(), ledP2z->text().toDouble());
     K2dBorderLine* hulpLine = new K2dBorderLine( a , b);
     //now we add the trianglesized
     hulpLine->changeTriaSize(ledBeginTriaSize->text().toDouble(), K2dBorderLine::BeginPoint);
     hulpLine->changeTriaSize(ledEndTriaSize->text().toDouble(), K2dBorderLine::EndPoint);
     //now we set the stuff off the base class, with the specific function written for that
     borderBaseSave(hulpLine);
     //add the line to the domain
     return dlgDomain->addBorder(index-1, hulpLine);
    }
  if (rbtSinCurve->isChecked())
    {
     K2dBorderSinCurve* hulpSine = new K2dBorderSinCurve( led_L->text().toDouble() ,
                     led_a->text().toDouble(),  led_V->text().toDouble(),  led_n->text().toUInt(),  led_b->text().toDouble() );
     //now we add the trianglesize
     hulpSine->changeTriaSize(ledSinTriangle->text().toDouble() );
     //set other values
     hulpSine->setReverse(chbReverse->isChecked()) ;
     hulpSine->setNrMeshInterval(ledSinMeshInt->text().toInt() );
     //now we set the stuff off the base class, with the specific function written for that
     borderBaseSave(hulpSine);
     //add the line to the domain
     return dlgDomain->addBorder(index-1, hulpSine);
    }
}
/** We need to seperate the code to save borderbase specific data
         A specific border (line, curve) calls this function to save this data in a general fashion*/
void dlgBorder::borderBaseSave(K2dBorderBase *hulpBorder)
{
  //we save the data of the bordercondition
  if (rbtNormal->isChecked())
     { hulpBorder->setBorderCond(K2dBorderBase::NEUMANN);  //it's a neumann cond, no other data since it's the only one possible
     } else {
         if (rbtDirichlet->isChecked() && rbtZ->isChecked() )
           { hulpBorder->setBorderCond(K2dBorderBase::DIRICHLET, K2dBorderBase::DirZ);
           } else {
             hulpBorder->setBorderCond(K2dBorderBase::DIRICHLET, K2dBorderBase::DirAx_B,
                                 ledA->text().toDouble(), ledB->text().toDouble() );
           }
     }

}
void dlgBorder::slotSaveExit()
{
  bool check = slotBorderSave();
  if (check)
    { //we say the dialog has been accepted
     this->accept();
    }
}
void dlgBorder::slotPreviousBorder()
{
  uint index=lcdNr->intValue();
  if (index == 1 )
     { //we are at beginning
       return;
     }
  --index;
  loadBorder(index);
}
void dlgBorder::slotNextBorder()
{
  uint index=lcdNr->intValue();
  if (index >= dlgDomain->nrOfBorders()  )
     { //we are at end
       return;
     }
  ++index;
  loadBorder(index);
}
void dlgBorder::slotAddEnd()
{
  //set index equal to number of Borders +1, and load that unexsisting border (=init values)
  uint index = dlgDomain->nrOfBorders() + 1;
  loadBorder(index);
  //since normally we make a closed domain, if we make a line, the beginning values must equal the previous last one
  QListIterator<K2dBorderBase> iterDom =  dlgDomain->iterator();
  //make iterator point at last border, so index - 2 position
   for (uint i=0 ; i < (index - 2)  ; ++i )
          { ++iterDom ;
          }
   QString iStr;
    ledP1x->setText(iStr.setNum(iterDom.current()->lastPoint()->coord1() ));
    ledP1z->setText(iStr.setNum( iterDom.current()->lastPoint()->coord2() ));
    ledBeginTriaSize->setText(iStr.setNum( iterDom.current()->lastPoint()->triaSize() ));
}
void dlgBorder::slotDelete()
{
  //first check if the border excists in the first place, if not, return
  uint index=lcdNr->intValue();
  if (index > dlgDomain->nrOfBorders()  )
     { //the border does not excist yet, so we can not delete it
       QMessageBox::information( this, "K2dGrFlow",
                            i18n("Border does not exist yet") );
       return;
     }
  if ( index <= dlgDomain->nrOfBorders() )
     { dlgDomain->removeBorder(index-1);   //the value in the list starts with 1
     }
  //now load the previous border on the screen, if there is one
  if (index > 1)
    { loadBorder(index-1);
    } else {
      loadBorder(1);
    }
}
