/***************************************************************************
                          domainnet2d.cpp  -  description
                             -------------------
    begin                : Wed May 30 2001
    copyright            : (C) 2001 by benny
    email                : bm@cage.rug.ac.be
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
//KDE includes
#include <kurl.h>
#include <qfile.h>
#include <qtextstream.h>

#include "domainnet2d.h"
//topology implementation
#include "topoimp.h"

DomainNet2d::DomainNet2d()
{  //makes empty vectors.
}
DomainNet2d::~DomainNet2d()
{ //destroy memory held by the domain
  //don't destroy the boundary, it was passed by a controller objected who has the responsilbility
  int i,sed(geoEdges.size()) , sel(geoEle.size()) , sn(geoNodes.size());
  for (i=0; i<sel; ++i)
       { delete geoEle[i]; };
  for (i=0; i<sed; ++i)
       { delete geoEdges[i]; };
  for (i=0; i<sn; ++i)
       { delete geoNodes[i] ;};
}
//we construct the entire net from inputfiles
void DomainNet2d::buildNet(KURL &EasyMeshFileName)
{
       int nrNodes, nrEdges, nrEle, nrBoundary(geoBoundary.size()) ;
       //dummy string to read stuff from the input
       QString strDummy ;
       //data we read in from the files
       int marker;
       double coord1,coord2;
       int nodei, nodej, nodek, eli, elj, elk, edgei, edgej, edgek ;
       //storage of results
       K2dBorderBase::BorderCondition bc, bc2;
       bool found, corner;
       int startBord, i, j, n, previousI;

       QString basePath = EasyMeshFileName.directory() + "/" + EasyMeshFileName.fileName() ;  // eg /home/me/k2dgrflow/test1.d
       basePath.remove(basePath.length()-2,2);           // remove the .d extension : position length-2, remove string of 2 long
       // put a textstream on the tree files to read
       QFile *fn = new QFile( basePath + ".n" );
       QFile *fs = new QFile( basePath + ".s" );
       QFile *fe = new QFile( basePath + ".e" );
       fn->open( IO_ReadOnly );
       QTextStream tn( fn );
       fs->open( IO_ReadOnly );
       QTextStream ts( fs );
       fe->open( IO_ReadOnly );
       QTextStream te( fe );
       //on the first line we find the number of nodes, sides (edges), elements
       tn >> nrNodes ;
       ts >> nrEdges ;
       te >> nrEle ;
       geoNodes.reserve(nrNodes);
       for( n=0; n<nrNodes; n++)
        {
         tn >> strDummy  >> coord1 >> coord2 >> marker ;
         if (marker > 0 )
           {
                 found = false ;
                 corner = false ;
                 //we want to start with the marker, because if done well, the marker will correspond with the boundary. StartBord is that.
                 if (marker <= nrBoundary) {startBord = marker-1 ;} else {startBord = 0 ;};
                 for (j=0 ;  (j < nrBoundary && (! found ) ) ; ++j)
                    {i = (startBord + j) % nrBoundary;
                     if (geoBoundary[ i]->onBorder(coord1,coord2,true))
                       { //see what border that is
                         found = true ;
                         if (! corner)
                            { corner = geoBoundary[ i]->endPoint(coord1,coord2) ;
                              bc = geoBoundary[ i]->getBorderCondition();
                              if (corner )
                                { //first time we see it's a corner, don't store the node yet, but look for the other boundary it's part of, set found again false
                                  previousI=i;     //remember the number of this border
                                  found = false ;
                                }
                                else
                                { // no corner, store the boundary node with the type of boundary it's on
                                  if (bc == K2dBorderBase::DIRICHLET)
                                    { geoNodes.push_back(new TopoNodeDir2d(coord1,coord2, i, false ) );
                                    } else if (bc == K2dBorderBase::NEUMANN)
                                    { geoNodes.push_back(new TopoNodeNeu2d(coord1,coord2, i, false ) );
                                    } else if (bc == K2dBorderBase::ROBIN)
                                    { geoNodes.push_back(new TopoNodeRobin2d(coord1,coord2, i, false ) );
                                    } else {
                                      cerr << "no correct boundary given, boundary point but did not find the border condition !\n" ;
                                    }
                                }
                            }
                            else
                            { //second time we find the cornernode, decide type of corner. Priority : dirichlet > neumann > robin
                              //bc contains bordercondition of the first border, look for the one of the second border
                              bc2 = geoBoundary[ i]->getBorderCondition();
                              //if corner node is on dirichlet ==> dirichletnode, else see if a neumann present, else see if robin present .
                              if ((bc == K2dBorderBase::DIRICHLET) || (bc2 ==K2dBorderBase::DIRICHLET) )
                                    { if (bc2 == K2dBorderBase::DIRICHLET)
                                        { geoNodes.push_back(new TopoNodeDir2d(coord1,coord2, i, true ) );
                                        } else {
                                          geoNodes.push_back(new TopoNodeDir2d(coord1,coord2, previousI , true ) );
                                        }
                                    } else if ( (bc == K2dBorderBase::NEUMANN) || (bc2 == K2dBorderBase::NEUMANN) )
                                    { if (bc2 == K2dBorderBase::NEUMANN)
                                        { geoNodes.push_back(new TopoNodeNeu2d(coord1,coord2, i, true ) );
                                        } else {
                                          geoNodes.push_back(new TopoNodeNeu2d(coord1,coord2, previousI , true ) );
                                        }
                                    } else if ( (bc == K2dBorderBase::ROBIN) || (bc2 == K2dBorderBase::ROBIN) )
                                    {  if (bc2 == K2dBorderBase::ROBIN)
                                         { geoNodes.push_back(new TopoNodeRobin2d(coord1,coord2, i , true ) );
                                         } else {
                                           geoNodes.push_back(new TopoNodeRobin2d(coord1,coord2, previousI , true ) );
                                         }
                                    } else {
                                      cerr << "no correct boundary given, boundary corner point but did not find the border condition !\n" ;
                                    }
                            }
                       } //end of if : looking on boundary for cycle looking for boundary
                    } //end of for cycle looking for boundary
                 if (!found)
                    {
                      cerr << "boundary point ( corner "<< corner << " ) but did not find associated border(s) .\n";
                    }
           } else {
             //no marker, so an inner node !
             geoNodes.push_back(new TopoNodeInner2d(coord1,coord2) ) ;
           }
        }
       fn->close();
       // finished reading in the nodes
       delete fn;
       //now we read the edge file a first time
       geoEdges.reserve(nrEdges);
       for(n=0; n<nrEdges; n++)
        {
         ts >> strDummy  >> nodei >> nodej >> eli >> elj >> marker ;
         //we can't use eli,j yet. If eli or elj is -1, we are on a border !
         // also, easymesh always makes lines as edges !
         if (eli == -1 || elj == -1)
           { //borderedge
             geoEdges.push_back(new TopoLineEdge2d(geoNodes[nodei], geoNodes[nodej], false));
           } else {
             geoEdges.push_back(new TopoLineEdge2d(geoNodes[nodei], geoNodes[nodej], true));
           }
        }
       fs->close();
       delete fs;
       //now we read the element file in a first time
       geoEle.reserve(nrEle);
       for(n=0; n<nrEle; n++)
        {
         te >> strDummy  >> nodei >> nodej >> nodek >> eli >> elj >> elk >> edgei >> edgej >> edgek >> coord1 >> coord2 >> marker ;
         //we can't use eli,j,k yet. Coord1 and 2, the coordinates of the circumcenter is not necessary for our use. Marker is not usefull for us.
         //we now easymesh makes triangles !
         //beware, we take : counterclockwise of local node 1 is local edge 1. This differs from the easymesh numbering !
         geoEle.push_back(new TopoTriangle2d( geoNodes[nodei ] , geoNodes[nodej], geoNodes[nodek],
                                                 (TopoLineEdge2d *)geoEdges[edgek],  (TopoLineEdge2d *)geoEdges[edgei],  (TopoLineEdge2d *)geoEdges[edgej] ) );
        }
       fe->close();
       //now we read the element file a second time to set element pointers correct
       fe->open(IO_ReadOnly);
       te >> nrEle ;
       for(n=0; n<nrEle; n++)
        {
         te >> strDummy  >> nodei >> nodej >> nodek >> eli >> elj >> elk >> edgei >> edgej >> edgek >> coord1 >> coord2 >> marker ;
         if (eli != -1) { geoEle[ n]->addNeighbour(geoEle[eli]); }
         if (elj != -1) { geoEle[ n]->addNeighbour(geoEle[elj]); }
         if (elk != -1) { geoEle[ n]->addNeighbour(geoEle[elk]); }
        }
       fe->close();
       delete fe;

}
bool DomainNet2d::testDomainNet2d()
{  //first test a triangle to see if that's ok
   TopoTriangle2d::testTopoTriangle2d();
   //now we make a domain
   cerr <<"constructing boundaries"  ;
   BorderPoint *b1 = new BorderPoint(0. , 0.);
   BorderPoint *b2= new BorderPoint(0. ,-1.) ;
   BorderPoint *b3= new BorderPoint(1.,0.) ;
   this->addBoundary(new K2dBorderLine(*b1 ,*b2  ) ) ;
   this->addBoundary(new K2dBorderLine(*b2 ,*b3 )  );
   this->addBoundary(new K2dBorderSinCurve(1. , 0. ,0.01,1,0.)) ;
   cerr << " .. ok \n" <<   "constructing nodes ";
   geoNodes.push_back( new TopoNodeDir2d(0.0,0.0,0,true) );    //0
   geoNodes.push_back( new TopoNodeNeu2d(0.0,-1.0, 1,true ) );  //1
   geoNodes.push_back( new TopoNodeDir2d(1.0, 0.0, 1, true) );  //2
   geoNodes.push_back( new TopoNodeDir2d(.75, 0.1,2, false) );  //3
   geoNodes.push_back( new TopoNodeDir2d(0.5, 0.0, 2, false) );  //4
   cerr << ".. ok \nconstructing edges" ;
   geoEdges.push_back( new TopoLineEdge2d(geoNodes[0],geoNodes[1]));   //0
   geoEdges.push_back( new TopoLineEdge2d(geoNodes[1],geoNodes[4]));   //1
   geoEdges.push_back( new TopoLineEdge2d(geoNodes[4],geoNodes[0]));   //2
   geoEdges.push_back( new TopoLineEdge2d(geoNodes[3],geoNodes[4]));   //3
   geoEdges.push_back( new TopoLineEdge2d(geoNodes[1],geoNodes[3]));   //4
   geoEdges.push_back( new TopoLineEdge2d(geoNodes[1],geoNodes[2]));   //5
   geoEdges.push_back( new TopoLineEdge2d(geoNodes[2],geoNodes[3]));   //6
   cerr <<" .. ok \nconstructing triangles" ;
   geoEle.push_back(new TopoTriangle2d(geoNodes[0],geoNodes[1], geoNodes[4],(TopoLineEdge2d*)geoEdges[0],(TopoLineEdge2d*)geoEdges[1],(TopoLineEdge2d*)geoEdges[2]));
   geoEle.push_back(new TopoTriangle2d(geoNodes[4],geoNodes[1], geoNodes[3],(TopoLineEdge2d*)geoEdges[1],(TopoLineEdge2d*)geoEdges[4],(TopoLineEdge2d*)geoEdges[3]));
   geoEle.push_back(new TopoTriangle2d(geoNodes[3],geoNodes[1], geoNodes[2],(TopoLineEdge2d*)geoEdges[4],(TopoLineEdge2d*)geoEdges[5],(TopoLineEdge2d*)geoEdges[6]));
   cerr <<"..ok\n" ;
   geoBoundary[2]->setBorderCond(K2dBorderBase::DIRICHLET, K2dBorderBase::DirZ) ;
   cerr << geoBoundary[0]->onBorder(geoNodes[2]->co1(),geoNodes[2]->co2(), false) << " : should be false,  node on line  \n" ;
   cerr << geoBoundary[0]->onBorder(geoNodes[1]->co1(),geoNodes[1]->co2(), false) << " : should be true \n" ;
   cerr << geoBoundary[1]->onBorder(geoNodes[1]->co1(),geoNodes[1]->co2(), false) << " : should be true \n" ;
   cerr << geoBoundary[1]->onBorder(geoNodes[4]->co1(),geoNodes[4]->co2(), false) << " : should be false\n" ;
   cerr << geoBoundary[2]->onBorder(geoNodes[4]->co1(),geoNodes[4]->co2(), false) << " : should be true, node on sine  \n" ;
   cerr << geoBoundary[2]->onBorder(0.0001, 0. , false) << " : should be false \n" ;
   cerr << geoBoundary[2]->endPoint(geoNodes[2]->co1(),geoNodes[2]->co2()) << " : should be true, endpoint  \n" ;
   cerr << geoBoundary[2]->gDirichlet(geoNodes[3]->co1(),geoNodes[3]->co2()) << " : should be 0.1, dirichlet value \n" ;
   geoBoundary[2]->setBorderCond(K2dBorderBase::DIRICHLET, K2dBorderBase::DirAx_B, 0.1, 0.001) ;
   cerr << geoBoundary[2]->gDirichlet(geoNodes[3]->co1(),geoNodes[3]->co2()) << " : should be 0.1*0.75+0.001=0.076, dirichlet value \n" ;
   return true;
}
void DomainNet2d::output()
{
  int i;
  int nrNodes(geoNodes.size()) ;
  for ( i=0 ; i < nrNodes ; ++i )
      { cout << "node " << i << " ( " << geoNodes[i]->co1() <<" ," <<geoNodes[i]->co2() << " ), inner " << geoNodes[i]->innerNode() ;
        cout << ", dirichlet " << geoNodes[i]->dirNode()  << ", neumann " << geoNodes[i]->neuNode() << ", corner " << geoNodes[i]->cornerNode() <<"\n";
      }
  int nrEdges(geoEdges.size()) ;
  for ( i=0 ; i < nrEdges ; ++i )
      { cout<< "edge " << i << " , inner " << geoEdges[i]->innerEdge() <<"\n";
      }
  int nrEle(geoEle.size()) ;
  for ( i=0 ; i < nrEle ; ++i )
      { cout<< "element " << i << " , surface " << geoEle[i]->surface() <<"\n";
      }
}
