/***************************************************************************
                          galarimpgrflow.cpp  -  description
                             -------------------
    begin                : Tue Jun 5 2001
    copyright            : (C) 2001 by benny
    email                : bm@cage.rug.ac.be
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "galarimpgrflow.h"
#include "topodefs.h"
#include "domainnet2d.h"
#include "topoimp.h"

#include <fstream.h>

#ifndef ID_STATUS_MSG
#define ID_STATUS_MSG 1
#endif

VEC* GalarImpGrFlow::solution()
{ //first, we calculate the local matrices of the fresh elements.
  cout << "Constructing local matrices...\n" ;
  int nrEle(net->nrEle());
  for (int iEle = 0 ; iEle < nrEle ; ++iEle)
     { // if fresh, calculate local matrix
       if (net->geoEle[iEle]->freshLocMat() )
         { localStiffness(net->geoEle[iEle] );
           localForce(net->geoEle[iEle]);
           //now the element is not fresh anymore
           net->geoEle[iEle]->setFresh(false);
         }
     }
  //now we assemble the global matrices out of the local (inherited methods of parent)
  cout << "Assembling the global Stiffness matrix...\n";
  assembleStiffness();
  // test ***********************************************
  //printStiffness("/home/benny/stiffness.m");
  // test ***********************************************
  cout << "Assembling the global Force vector...\n";
  assembleForce();
  // test ***********************************************
  //printForce("/home/benny/force.m");
  // test ***********************************************
  //we solve stiffness * unknowns = force

  //step one : a seperate matrix for the factorisation
  cout << "Doing Incomplete Cholesky factorisation of the stiffness matrix...\n";
  SPMAT* LLT = sp_copy(stiffness);
  spICHfactor(LLT);
  //step two, solving with iteration steps.
  cout << "Solving to unknowns by iteration...\n";
  int num_steps;
  iter_spcg(stiffness, LLT, force, 1e-8, unknowns, 2000, &num_steps ); // precision 1e-8, maximum of 2000 iteration steps
  cout << "Solved in " << num_steps << " iteration steps. \n" ;
  return unknowns;
}
/** calculate the localstiffness matrix with a specific element */
void GalarImpGrFlow::localStiffness(TopoElement2d* el)
{ //we now it's a triangle in this implementation, so no checking on element type,
  //   we know 3 local nodes, and we now only a diffusion term
  double test;
  for (uint i=0 ; i< 3 ; ++i)
    { for (uint j=0 ; j < 3 ; ++j)
        {  el->setLocMat(i,j, diffusion(el, i , j) ) ;
        }
    }
}
/** calculate the local force vector with a specific element */
void GalarImpGrFlow::localForce(TopoElement2d* el)
{ //we now it's a triangle in this implementation, so no checking on element type,
  //   we know 3 local nodes, and we now only a source term
  for (uint i=0 ; i< 3 ; ++i)
    { el->setLocForce(i, sourceTerm(el, i)) ;
    }
}
/** calculate the diffusion term for local node I,J in an element */
double GalarImpGrFlow::diffusion(TopoElement2d* tria, unsigned int localI, unsigned int localJ)
{//when K_i_i is asked and node dirichlet, then return 1, else calculate
  //TopoTriangle2d *tria = (TopoTriangle2d*) el;     // no casting not necessary
  if (localI == localJ)
    { if (tria->node(localI)->dirNode())
        { return 1.;
        } else {
          //return kii for non dirichlet, depending on d !
          int j((localI+1)%3) , k ((localI+2)%3) ;
          if (d==0.)
            { return (1./ (4. * tria->surface()) * ( (tria->node( j)->co2() - tria->node(k)->co2())
                                                                      * (tria->node( j)->co2() - tria->node(k)->co2())
                                                                  + (tria->node( j)->co1() - tria->node(k)->co1())
                                                                      * (tria->node( j)->co1() - tria->node(k)->co1())  ) );
            } else if (tria->node(j)->co2() == tria->node(k)->co2()
                               || tria->node(localI)->co2() == tria->node(j)->co2()
                               || tria->node(localI)->co2() == tria->node(k)->co2() )
            { //denuminator of normal formula is 0, this is rare !!!
              //determine equals, and the nonequal, and use formule you get with rule of l'hopital
              int equal, nonequal;
              if (tria->node(j)->co2() == tria->node(k)->co2() ) {equal = j; nonequal = localI;}
              if (tria->node(localI)->co2() == tria->node(j)->co2() ) {equal = j; nonequal = k;}
              if (tria->node(localI)->co2() == tria->node(k)->co2() ) {equal = localI; nonequal = j;}
              return ( 1./ (2. * tria->surface()) * ( (tria->node( j)->co2() - tria->node(k)->co2())
                                                                       * (tria->node( j)->co2() - tria->node(k)->co2())
                                                                  + (tria->node( j)->co1() - tria->node(k)->co1())
                                                                       * (tria->node( j)->co1() - tria->node(k)->co1()) )
                            * (exp(d* tria->node(nonequal)->co2()) - exp(d* tria->node(equal)->co2())
                                 + exp(d* tria->node(equal)->co2()) *d * (tria->node(equal)->co2()-tria->node(nonequal)->co2()) )
                            / (d*d*(tria->node(equal)->co2()-tria->node(nonequal)->co2())
                                     *(tria->node(equal)->co2()-tria->node(nonequal)->co2()) )
                         );
            } else {
                 return (1./ (2. * tria->surface()) * ( (tria->node( j)->co2() - tria->node(k)->co2())
                                                                       * (tria->node( j)->co2() - tria->node(k)->co2())
                                                                  + (tria->node( j)->co1() - tria->node(k)->co1())
                                                                       * (tria->node( j)->co1() - tria->node(k)->co1()) )
                           * ( (tria->node(j)->co2() - tria->node(k)->co2() ) * exp(d * tria->node(localI)->co2()) +
                                (tria->node(k)->co2() - tria->node(localI)->co2() ) * exp(d * tria->node(j)->co2()) +
                                 (tria->node(localI)->co2() - tria->node(j)->co2() ) * exp(d * tria->node(k)->co2())  )
                           / (d * d *( tria->node(j)->co2() - tria->node(k)->co2() )* (tria->node(localI)->co2() - tria->node(j)->co2())
                                       * (tria->node(localI)->co2() - tria->node(k)->co2() ) ) ) ;
            }
        }
    } else {
      //determine the value of the tirth node k. Beware, i j k is not in counterclockwise order. Not necessary for our formulas
      int k(3-localI-localJ);
      if (d==0.)
        { //kij
          return ( 1./ (4. * tria->surface()) * ( (tria->node(k)->co2() - tria->node(localI)->co2() )
                                                                 *  (tria->node(localJ)->co2() - tria->node(k)->co2() )
                                                               +(tria->node(k)->co1() - tria->node(localI)->co1() )
                                                                 *  (tria->node(localJ)->co1() - tria->node(k)->co1() )  )  );
        } else if (tria->node(localJ)->co2() == tria->node(k)->co2()
                               || tria->node(localI)->co2() == tria->node(localJ)->co2()
                               || tria->node(localI)->co2() == tria->node(k)->co2() )
        { //denuminator of normal formula is 0, this is rare !!!
          //determine equals, and the nonequal, and use formule you get with rule of l'hopital
          int equal, nonequal;
          if (tria->node(localJ)->co2() == tria->node(k)->co2() ) {equal = k; nonequal = localI;}
          if (tria->node(localI)->co2() == tria->node(localJ)->co2() ) {equal = localI; nonequal = k;}
          if (tria->node(localI)->co2() == tria->node(k)->co2() ) {equal = localI; nonequal = localJ;}
          return  ( 1./ (2. * tria->surface()) * ( (tria->node(k)->co2() - tria->node(localI)->co2() )
                                                                 *  (tria->node(localJ)->co2() - tria->node(k)->co2() )
                                                               +(tria->node(k)->co1() - tria->node(localI)->co1() )
                                                                 *  (tria->node(localJ)->co1() - tria->node(k)->co1() )  )
                         * (exp(d* tria->node(nonequal)->co2()) - exp(d* tria->node(equal)->co2())
                              + exp(d* tria->node(equal)->co2()) *d * (tria->node(equal)->co2()-tria->node(nonequal)->co2()) )
                         / (d*d*(tria->node(equal)->co2()-tria->node(nonequal)->co2())
                                  *(tria->node(equal)->co2()-tria->node(nonequal)->co2()) )
                      );
        } else {
          return ( 1./ (2. * tria->surface()) * ( (tria->node(k)->co2() - tria->node(localI)->co2() )
                                                                 *  (tria->node(localJ)->co2() - tria->node(k)->co2() )
                                                               +(tria->node(k)->co1() - tria->node(localI)->co1() )
                                                                 *  (tria->node(localJ)->co1() - tria->node(k)->co1() )  )
                           * ( (tria->node(localJ)->co2() - tria->node(k)->co2() ) * exp(d * tria->node(localI)->co2()) +
                                (tria->node(k)->co2() - tria->node(localI)->co2() ) * exp(d * tria->node(localJ)->co2()) +
                                 (tria->node(localI)->co2() - tria->node(localJ)->co2() ) * exp(d * tria->node(k)->co2())  )
                           / (d*d * (tria->node(localJ)->co2() - tria->node(k)->co2())
                                    * (tria->node(localI)->co2() - tria->node(localJ)->co2())
                                    * (tria->node(localI)->co2() - tria->node(k)->co2()) )
                     ) ;
        }
    }
}
