/***************************************************************************
                          galarsolver2d.cpp  -  description
                             -------------------
    begin                : Tue May 29 2001
    copyright            : (C) 2001 by benny
    email                : bm@cage.rug.ac.be
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
#include <math.h>
#include <fstream.h>

#include "galarsolver2d.h"
#include "topodefs.h"
#include "domainnet2d.h"
#include "topology.h"
#include "topoimp.h"


GalarSolver2d::GalarSolver2d(DomainNet2d* domnet) : net(domnet)
{   //for every node a solution must be found
    int nr(domnet->nrNodes());
    stiffness = sp_get( nr, nr , MAX_ELE_PER_NODE);  //matrix nr x nr, a row has never more than MAX_ELE_PER_NODE non zero values
    force = v_get( nr);
    unknowns = v_get( nr);
}
GalarSolver2d::~GalarSolver2d(){
    //free the space of the matrices
    SP_FREE(stiffness);
    V_FREE(force);
    V_FREE(unknowns);
}
SPMAT*  GalarSolver2d::assembleStiffness()
{ //assemble the stiffness matrix out of the local matrices

  //we start assembling
  int nr(net->nrNodes());
  int i(0), j, si, k, locali, localj;
  for ( ; i < nr ; ++i)
   { for (j=0; j < nr; ++j)
       { //we make stiffness matrix K_i_j
        //how many elements is i part of?
        si = net->geoNodes[i]->inEleNr();
        //for every element where i is in, see if j is in it too
        if (i==j)
          {  // K_i_i value, just run over all the elements it's part of
             for (k=0 ; k < si ; ++k)
              { //see what the locali parameter is (which MUST excist), as long as the node in the element is not the same
                // as the node we start with, we have not found the locali
                for ( locali=0; net->geoNodes[i]->eleNodePartOf( k)->node(locali) != net->geoNodes[i] ; ++locali){} ;
                //now locali is known, we get the local matrix element local i , local i and put it in the global;
                //extra : if 0, we add also, therefore a compact at the end !
                 sp_set_val_dir(stiffness , i,i,
                        sp_get_val_dir(stiffness,i,i) + net->geoNodes[i]->eleNodePartOf(k)->locMat((uint)locali, (uint)locali) );
              }
          } else {
            for (k=0 ; k < si ; ++k)
              { //check for every element if j is also a node of that element
                localj = net->geoNodes[j]->inElement(net->geoNodes[i]->eleNodePartOf(k));
                if (localj != -1 )
                  { //ok we know the localj parameter in the element k where i is part of.
                    //see what the locali parameter is (which MUST excist)
                    for ( locali=0; net->geoNodes[i]->eleNodePartOf( k)->node(locali) != net->geoNodes[i] ; ++locali){} ;
                    //now locali is also known, we get the local matrix element local i , local j and put it in the global;
                    //extra : if 0, we add also, therefore a compact at the end !
                    sp_set_val_dir(stiffness , i, j,
                             sp_get_val_dir(stiffness,i,j) + net->geoNodes[i]->eleNodePartOf(k)->locMat((uint)locali, (uint)localj) );
#ifdef TOPOCHECK
                    //cout << "K_ij, " << i << " " << j << ", element " << k << " local nodes " << locali << " " << localj << " ,value: "
                    //        << sp_get_val(stiffness,i,j) << "\n";
#endif
                  }
                //goto next element i is part of
              }
          }
        //goto next ij pair
       }
   }
  //now we remove the too small values out of our sparse matrix, and also the 0 values
  //stiffness = sp_compact(stiffness , 0.0);
  return stiffness;
}
VEC*  GalarSolver2d::assembleForce()
{ //assemble the force vector matrix out of the local forces

  //we start assembling
  int nr(net->nrNodes());
  int j(0), sj, k(0), localj;
  //global numbering
  for ( ; j < nr ; ++j)
   { //determine for every element node j is part of, what the local number is, and get the force associated with that number
     //how many elements is node j part of?
     sj = net->geoNodes[j]->inEleNr();
     //run over those elements
     for (k=0 ; k < sj ; ++k)
       { //determine local j
         for ( localj=0; net->geoNodes[ j]->eleNodePartOf( k)->node(localj) != net->geoNodes[j] ; ++localj){} ;
         //now locali is known, we get the local vectorforce element local j , local i and put it in the global force vector;
         force->ve[ j] = force->ve[ j] +
                                   net->geoNodes[ j]->eleNodePartOf(k)->locForce((uint)localj) ;
#ifdef TOPOCHECK
         //cout << "F_j, " << j << ", element " << k << " local node " << localj << " ,value: "
         //        << force->ve[ j] << "\n";
#endif
       }
   }
  return force;
}
void GalarSolver2d::initStiffness()
{  //quickest : just free and reallocate
  SP_FREE(stiffness);
  int nr(net->nrNodes());
  stiffness = sp_get( nr, nr , MAX_ELE_PER_NODE);  //matrix nr x nr, a row has never more than MAX_ELE_PER_NODE non zero values
}
void GalarSolver2d::initForce()
{ //quickest : just free and reallocate
  V_FREE(force);
  int nr(net->nrNodes());
  force = v_get( nr);
}
void GalarSolver2d::printStiffness(const char *filename)
{
  FILE *foutputFil;
  foutputFil = fopen(filename,"w");
  sp_foutput(foutputFil, stiffness);
  fclose(foutputFil);
}
void GalarSolver2d::printForce(const char *filename)
{
  FILE *foutputFil;
  foutputFil = fopen(filename,"w");
  v_foutput(foutputFil, force);
  fclose(foutputFil);
}
void GalarSolver2d::printSolFile(const char *filename)
{
  FILE *foutputFil;
  foutputFil = fopen(filename,"w");
  v_foutput(foutputFil, unknowns);
  fclose(foutputFil);
}
/* a method to read in vectors from file without checking the size
   We use this instead of the version in the meschach library : v_finput */
VEC * GalarSolver2d::readVEC(FILE *fp, VEC *vec)
{
     u_int      i,dim;
     int        io_code;

     /* get dimension */
     skipjunk(fp);
     //change with original of library : no check on maximum size !
     if ((io_code=fscanf(fp," Vector: dim:%u",&dim)) < 1 )
	  error(io_code==EOF ? 7 : 6,"bfin_vec");

     /* allocate memory if necessary */
     if ( vec==(VEC *)NULL )
	  vec = v_resize(vec,dim);

     /* get entries */
     skipjunk(fp);
     for ( i=0; i<dim; i++ )
	  if ((io_code=fscanf(fp,"%lf",&vec->ve[i])) < 1 )
	       error(io_code==EOF ? 7 : 6,"bfin_vec");

     return (vec);
}
/* sp_set_val _dir_expand-- sets the (i,j) entry of the sparse matrix A after expanding it*
    is an intagral part of the sp_set_val_dir funtion that is inlined ! */
double GalarSolver2d::sp_set_val_dir_expand(SPMAT *A, int i,int j, double val, SPROW * r, int idx)
{
   int	idx2, new_len;

   if ( idx < -1 )
   {
      /* Note: this destroys the column & diag access paths */
      A->flag_col = A->flag_diag = FALSE;
      /* shift & insert new value */
      idx = -(idx+2);	/* this is the intended insertion index */
      if ( r->len >= r->maxlen )
      {
	 r->len = r->maxlen;
	 new_len = max(2*r->maxlen+1,5);
	 if (mem_info_is_on()) {
	    mem_bytes(TYPE_SPMAT,A->row[i].maxlen*sizeof(row_elt),
			    new_len*sizeof(row_elt));
	 }

	 r->elt = RENEW(r->elt,new_len,row_elt);
	 if ( ! r->elt )	/* can't allocate */
	   error(E_MEM,"sp_set_val");
	 r->maxlen = 2*r->maxlen+1;
      }
      for ( idx2 = r->len-1; idx2 >= idx; idx2-- )
	MEM_COPY((char *)(&(r->elt[idx2])),
		 (char *)(&(r->elt[idx2+1])),sizeof(row_elt));
      /************************************************************
	if ( idx < r->len )
	MEM_COPY((char *)(&(r->elt[idx])),(char *)(&(r->elt[idx+1])),
	(r->len-idx)*sizeof(row_elt));
	************************************************************/
      r->len++;
      r->elt[idx].col = j;
      return r->elt[idx].val = val;
   }
   /* else -- idx == -1, error in index/matrix! */
   return 0.0;
}

