/***************************************************************************
                          k2dborderbase.cpp  -  description
                             -------------------
    begin                : Mon Apr 30 2001
    copyright            : (C) 2001 by benny
    email                : bm@cage.rug.ac.be
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
//General
#include <math.h>
#include <fstream.h>
//QT includes

//KDE includes
#include <klocale.h>     //language traduction

// application specific includes
#include "k2dborderbase.h"
#include "topodefs.h"

BorderPoint::BorderPoint(double coord1 , double coord2, double triaSize)
{
   data.resize(3);
   data[0]=coord1 ;
   data[1]=coord2 ;
   data[2]=triaSize ;
}
BorderPoint::BorderPoint(const BorderPoint &bp)
{
   data.resize(3);
   data.duplicate( bp.data) ;
}

K2dBorderBase::K2dBorderBase()
{
  borderCond = NotSet;
  // we initialize the other data, although they have no meaning if bordercond is not set yet.
  //  so users should first set them before asking them.
  dirType = DirZ;
  dirLineA = 0;
  dirLineB = 0;
  points = 0;
}
bool K2dBorderBase::endPoint(const double &co1, const double &co2)
{ //check the given coordinates with first and lastpoint.
  if ( (co1 == firstPoint()->coord1() && co2 == firstPoint()->coord2()) ||
        (co1 == lastPoint()->coord1() && co2 == lastPoint()->coord2())  )
     { return true ; }
   return false;
}
double K2dBorderBase::gDirichlet(const double &co1, const double &co2) const
{
   if (borderCond == DIRICHLET)
     { //valid question
       if (dirType == DirZ)
         {  //the dirichlet function is the second coordinate
            return co2 ;
         } else if (dirType == DirAx_B)
         { //a simple function depending on the first coordinate
           return dirLineA * co1 + dirLineB ;
         } else {
           cerr <<  "Dirichletvalue is undefined, errors will occur, nothing returned ";
         }
     } else {
       cerr << "Asked Dirichletvalue but no dirichlet border. You did something wrong ! " ;
     }
}
double K2dBorderBase::gNeumann(const double &co1, const double &co2) const
{
   if (borderCond == NEUMANN)
     { //valid question, only normal derivitive = 0 is possible for now, so return 0 as neumann function
       return 0.  ;
     } else {
       cerr << "Asked Neumanvalue but no neumann border. You did something wrong ! " ;
     }
}
/** construct empty line object, use this with care, since not initialized ! */
K2dBorderLine::K2dBorderLine() : K2dBorderBase()
{
  //a line consists of two points
  points.resize(2);
}

/**  construct a line given the begin and the end point */
K2dBorderLine::K2dBorderLine(const BorderPoint &beginpt, const BorderPoint &endpt) : K2dBorderBase()
{
  //a line consists of two points
  points.resize(2);
  points[0] = new BorderPoint(beginpt);
  points[1] = new BorderPoint(endpt);
}
K2dBorderLine::~K2dBorderLine(){
}
void K2dBorderLine::changePoint(BorderPoint &pt , LinePoint lp )
{
  if (lp == BeginPoint)
    { points[0]->duplicate(pt);
    };
  if (lp == EndPoint)
    { points[1]->duplicate(pt);
    };
}
bool K2dBorderLine::onBorder(const double &co1, const double &co2, bool meshed) const
{  // on the line through (x1,y1) and (x2,y2) if  (y - y1) - ((y2-y1)/(x2-x1)) (x - x1) = 0
   // for us = 0 is translated to fabs < TOPONUMEPS
   // furthermore x must be between x1 and x2
   double x1(firstPoint()->coord1()), x2(lastPoint()->coord1()), y1(firstPoint()->coord2()), y2(lastPoint()->coord2());
   if ( x1<= x2 )
     { if ( (x1 > co1) || ( x2< co1) ) {return false; }
     } else {
       if ( (x2 > co1) || (x1 < co1) ) {return false; }
     }
   // check if on the line
   if (x1 == x2)
     { if (y1<= y2)
         { return ( co2>=y1 ) && ( co2 <= y2 );
         } else {
           return (co2 >= y2) && ( co2 <= y1 );
         }
     } else {
       return  (fabs (co2-y1 - (y2 - y1) / (x2 - x1) * (co1 - x1)) < TOPOTRIAEPS ) ;
     }
   return false ;
}
void K2dBorderLine::minCo(uint coordIndex, double &valueMin) const
{
   //a line is determined by begin and end
   if (firstPoint()->at(coordIndex) < valueMin) {valueMin = firstPoint()->at(coordIndex); } ;
   if (lastPoint()->at(coordIndex) < valueMin) {valueMin = lastPoint()->at(coordIndex) ;} ;
}
void K2dBorderLine::maxCo(uint coordIndex, double &valueMax) const
{
   //a line is determined by begin and end
   if (firstPoint()->at(coordIndex) > valueMax) {valueMax = firstPoint()->at(coordIndex); } ;
   if (lastPoint()->at(coordIndex) > valueMax) {valueMax = lastPoint()->at(coordIndex) ; } ;
}
/**  Return trianglesize in begin or endpoint, is default initialized on construction by 1, so 1 is returned if
      never set before */
double K2dBorderLine::triaSize(LinePoint lp)
{
  if (lp == BeginPoint)
    { return points[0]->at(2);
    };
  if (lp == EndPoint)
    { return points[1]->at(2);
    };
}
/** Change the trianglesize in begin or endpoint */
void K2dBorderLine::changeTriaSize(double size, LinePoint lp)
{
  if (lp == BeginPoint)
    { points[0]->at(2) = size;
    };
  if (lp == EndPoint)
    { points[1]->at(2) = size ;
    };
}
/** draw on a painter the line */
void K2dBorderLine::draw(QPainter* p, bool meshed )
{ //nomatter the value of meshed, a line is always a line
    switch (borderCond )
       {case NEUMANN :
                 p->setPen(Qt::red);
                 break;
        case DIRICHLET :
                 p->setPen(Qt::blue);
                 break;
        default:
                 p->setPen(Qt::black);
                 break;
       };
    //function requires integers, we convert our doubles to that
    p->drawLine(points[0]->at(0),points[0]->at(1), points[1]->at(0), points[1]->at(1));
}

void K2dBorderLine::printBorder(QTextStream &os)
{
  os << *this ;
}
QTextStream &operator<<( QTextStream & os , const K2dBorderLine & line)
{
  os << i18n("Line from (")  ;
  os << line.firstPoint()->coord1() << ", " <<   line.firstPoint()->coord2() << i18n(" ) to (")      ;
  os << line.lastPoint()->coord1() << ", " <<   line.lastPoint()->coord2() ;
  os << i18n(") \n") ;
  return os;
}
K2dBorderSinCurve::K2dBorderSinCurve(double L, double a, double V, uint n, double b ) : K2dBorderBase()
{
  //set values
  amplitude = V;
  hills = n;
  length = L;
  offset = b;
  slope = a ;
  //no intervalpoints yet
  nrMeshPoints = 0;
  //normal order
  reverse = false ;
  //a sinus starts with two points, begin and end
  points.resize(2);
  points[0] = new BorderPoint( value1(0.0), value2(0.0), 1.0);
  points[1] = new BorderPoint( value1(1.0), value2(1.0), 1.0);
}
K2dBorderSinCurve::~K2dBorderSinCurve(){
}
/** draw on a painter the sinus */
void K2dBorderSinCurve::draw(QPainter* p , bool meshed )
{
  switch (borderCond )
       {case NEUMANN :
                 p->setPen(Qt::red);
                 break;
        case DIRICHLET :
                 p->setPen(Qt::blue);
                 break;
        default:
                 p->setPen(Qt::black);
                 break;
       };
  //draw the sinus
  //first determine the biggist scaling
  double maxScale = p->worldMatrix().m11();
  if (maxScale < p->worldMatrix().m22() ) {maxScale = p->worldMatrix().m22();}
  //decide interval with wich we will draw the sinus (every 5 pixels a line)
  double interval=1.0;
  if ( length != 0.0 &&  maxScale != 0.0 ) { interval = 5.0 / (length * maxScale); } ;
  for (double test = interval  ; test <= 1.0 ; test = test + interval )
    {
      p->drawLine(value1(test-interval) , value2(test-interval ) , value1(test) , value2(test) );
    }
  //check if we have to perform a last line
  if (test > 1.0) {p->drawLine(value1(test-interval) , value2(test-interval ) , value1(1.0) , value2(1.0) ); }

  //if necessary also the mesh approximation
  if (meshed)
    { p->setPen(Qt::green);
     for (int i = 0 ; i < points.size()-1 ; ++i)
        { //draw all the lines of the mesh
          p->drawLine(points[i]->at(0),points[i]->at(1), points[i+1]->at(0), points[i+1]->at(1));
        }
    }
}
/** Change the sine in one action. This automatically recalculates interval points
         One action for less overhead in recalculating interval */
void K2dBorderSinCurve::changeSinCurve( double L, double a, double V, uint n, double b )
{
  //set values
  amplitude = V;
  hills = n;
  length = L;
  offset = b;
  slope = a ;
  //we change begin and end point to the new values
  double tria = triaSize();
  BorderPoint* temp;
  temp = points[0];
  if (!reverse)
   {points[0] = new BorderPoint( value1(0.0), value2(0.0), tria); }
   else {
    points[0] = new BorderPoint( value1(1.0), value2(1.0), tria);
   }
  delete temp;
  temp = points[points.size()-1];
  if (!reverse)
     {points[points.size()-1] = new BorderPoint( value1(1.0), value2(1.0), tria ); }
     else {
     points[points.size()-1] = new BorderPoint( value1(0.0), value2(0.0), tria);
     }
  delete temp;
  //we need to redo the meshing intervals, so we call the function that does this
  calcMeshIntervalPoints();
}
bool K2dBorderSinCurve::onBorder(const double &co1, const double &co2, bool meshed) const
{//  if we check on the meshed border, we have to check each line seperately (see how it works for a line )
 //  if we check on the sinus, we see if point is on the sinus, and not outside the begin and end

 //check if not outside the total piece
 double x1(firstPoint()->coord1()), x2(lastPoint()->coord1());
 if ( x1< x2 )
   { if ( (x1 > co1) || ( x2< co1) ) {return false; }
   } else {
     if ( (x2 > co1) || (x1 < co1) ) {return false; }
   }
 if (meshed )
   {  // check if on a piece of the meshed sinus
      double  y1(points[0]->coord2()), y2;
      bool check;
      for (int i = 1; i < nrPointsOnBorder() ; ++i )
         { // see for every line if point is on it
           check = true ;
           x2 = points[ i]->coord1();
           y2 = points[ i]->coord2();
           //when x is between the borders of the cutted line, then ok
           if ( x1< x2 )
             { if ( (x1 > co1) || ( x2< co1) ) {check = false ; }
             } else {
               if ( (x2 > co1) || (x1 < co1) ) {check = false; }
             }
           //so only when check, we test this particular piece
           if (check )
             { //first see if not a vertical
               if (x1 == x2)      //vertical
                 { if (y1<= y2)
                     { if (( co2>=y1 ) && ( co2 <= y2 )) {return true ;};
                     } else {
                       if ((co2 >= y2) && ( co2 <= y1 )) {return true ;};
                     }
                 } else {   // a normal line
                   if (fabs (co2-y1 - (y2 - y1) / (x2 - x1) * (co1 - x1)) < TOPOTRIAEPS ) {return true;}
                 }
             }
           //prepare for next piece, end of this piece is begin of next
           x1 = x2;
           y1 = y2;
         }
   } else{
      //check if on the sinus, we now the co1 value is ok, we calculate the sinus with that value and compare with input
      // define parameter t
      double t( (co1 - offset) / length);
      //see if corresponding sine value is equal to co2 : abs(co2 - sinus ) < TOPONUMEPS
      if  ( fabs (-1.0 * ( (slope*t) + amplitude * sin(2.0 * M_PI * hills * t) ) ) < TOPOTRIAEPS ) {return true ;}
   }
   return false ;
}
/**we want to be able to retrieve the min and maximum values of border
         coordIndex is the index, eg 0 for the first, 1 for the second; and value is the present min or max
         the function changes this if necessary. */
void K2dBorderSinCurve::minCo(uint coordIndex, double &valueMin) const
{  // sinus in 1 direction, so for first x coord simpel
   if (coordIndex == 0)
      { if (firstPoint()->at(coordIndex) < valueMin) {valueMin = firstPoint()->at(coordIndex); } ;
        if (lastPoint()->at(coordIndex) < valueMin) {valueMin = lastPoint()->at(coordIndex) ;} ;
        return ;
      } else {
        double hulp;
        // coord 2 has a sinus. we not only have to look at begin and end, but also to crest
         if (firstPoint()->at(coordIndex) < lastPoint()->at(coordIndex) )
            { // first point lowest
              if (firstPoint()->at(coordIndex) < valueMin) {valueMin = firstPoint()->at(coordIndex); } ;
             //check if a hill of the sinus comes lower
              if (hills==0 || amplitude ==0 )
                 {// nothing to check
                   return ;
                 } else {
                   // mark that in our definition the real amplitude is the stored value x -1
                   if ( (-1.0 * amplitude) < 0.0 )
                     { //first bottom after t = 1/ 4n
                       hulp = value2( 1.0 / (4.0*hills) ) ;
                       if (hulp < valueMin) {valueMin = hulp; } ;
                     } else {
                       //first bottom after t = 3/ 4n
                       hulp = value2( 3.0 / (4.0*hills) ) ;
                       if (hulp < valueMin) {valueMin = hulp; } ;
                     }
                 }
            } else {
              // last point lowest
              if (lastPoint()->at(coordIndex) < valueMin) {valueMin = lastPoint()->at(coordIndex); } ;
             //check if a hill of the sinus comes lower
              if (hills==0 || amplitude ==0 )
                 {// nothing to check
                   return ;
                 } else {
                   if ( (-1.0 * amplitude) < 0.0 )
                     { //first bottom after t = 1-3/ 4n
                       hulp = value2( 1.0 - 3.0 / (4.0*hills) );
                       if (hulp < valueMin) {valueMin = hulp; } ;
                     } else {
                       //first bottom after t = 1- 1/ 4n
                       hulp = value2( 1.0 - 1.0 / (4.0*hills) );
                       if (hulp < valueMin) {valueMin = hulp; } ;
                     }
                 }
            }
      }
}
void K2dBorderSinCurve::maxCo(uint coordIndex, double &valueMax) const
{  // sinus in 1 direction, so for first coord simpel
   if (coordIndex == 0)
      { if (firstPoint()->at(coordIndex) > valueMax) {valueMax = firstPoint()->at(coordIndex); } ;
        if (lastPoint()->at(coordIndex) > valueMax) {valueMax = lastPoint()->at(coordIndex) ;} ;
        return ;
      } else {
        double hulp;
        // coord 2 has a sinus. we not only have to look at begin and end, but also to crest
         if (firstPoint()->at(coordIndex) > lastPoint()->at(coordIndex) )
            { // first point highest
              if (firstPoint()->at(coordIndex) > valueMax) {valueMax = firstPoint()->at(coordIndex); } ;
             //check if a hill of the sinus comes higher
              if (hills==0 || amplitude ==0 )
                 {// nothing to check
                   return ;
                 } else {
                   // mark that in our definition the real amplitude is the stored value times -1
                   if ( (-1.0 * amplitude) > 0.0 )
                     { //first height after t = 1/ 4n
                       hulp = value2( 1.0 / (4.0*hills) ) ;
                       if (hulp > valueMax) {valueMax = hulp; } ;
                     } else {
                       //first heigth after t = 3/ 4n
                       hulp = value2( 3.0 / (4.0*hills) ) ;
                       if (hulp > valueMax) {valueMax = hulp ;} ;
                     }
                 }
            } else {
              // last point highest
              if (lastPoint()->at(coordIndex) > valueMax) {valueMax = lastPoint()->at(coordIndex); } ;
             //check if a hill of the sinus comes higher
              if (hills==0 || amplitude ==0 )
                 {// nothing to check
                   return ;
                 } else {
                   if ( (-1.0 * amplitude) > 0.0 )
                     { //first heigth after t = 1-3/ 4n
                       hulp = value2( 1.0 - 3.0 / (4.0*hills) ) ;
                       if (hulp > valueMax) {valueMax = hulp; } ;
                     } else {
                       //first height after t = 1- 1/ 4n
                       hulp = value2( 1.0 - 1.0 / (4.0*hills) ) ;
                       if (hulp > valueMax) {valueMax = hulp; } ;
                     }
                 }
            }
      }
}
/** set the meshinterval to a certain value (default = 0), the points are recalculated if changed.
         beware to give value >=0 */
void K2dBorderSinCurve::setNrMeshInterval(int nr)
{
  if (nr == nrMeshPoints)
    { //nothing changes
      return ;
    } else {
      nrMeshPoints = nr;
      //we need to redo the meshing intervals, so we call the function that does this
      calcMeshIntervalPoints();
    }
}
/** set how the curve is transversed */
void K2dBorderSinCurve::setReverse(bool rev)
{
  if (reverse == rev)
    { //nothing changes
      return;
    } else {
      reverse = rev;
      BorderPoint *temp;
      int size = points.size();
      //reverse the order of points
      for (int i=1; i <= points.size() / 2; ++i )
        { temp = points[i-1];
          points[i-1] = points[size- i];
          points[size-i] = temp;
        }
    }
}
/**  Return trianglesize in meshpoints on the sine, is default initialized on construction by 1, so 1 is returned if
         never set before */
double K2dBorderSinCurve::triaSize() const
{ //triangle size constant over a sine, return the one of the first point
  return points[0]->at(2);
}
/** Change the trianglesize for the meshpoints */
void K2dBorderSinCurve::changeTriaSize(double size)
{ //since all points have the same triasize,this comes down to putting it the same as size everywhere
  for (int i = 0; i < points.size(); ++i)
     {  points[i]->setTriaSize(size);
     }
}
/** Implementation of base class printBorder*/
void K2dBorderSinCurve::printBorder(QTextStream &os)
{  //just pass to the << operator of the stream
  os << *this ;
}
/** Internal function to change the points datastructure to the desired number of points
    Every points structure already contains first and last point, rest is reset, size is altered */
void K2dBorderSinCurve::calcMeshIntervalPoints()
{ //first resize to 1 to free space contained by current points
  points.resize(1);
  //calculate the step size of parameter t from 0..1
  double stepSize = 1.0/ (nrMeshPoints + 1.0) ;
  double hulpTria = points[0]->triaSize();
  //make points sufficiently large, and add the points, but not yet the last one;
  points.resize(nrMeshPoints + 2);
  if (!reverse)
    {
     for (int i = 1; i < points.size()-1; ++i)
        {  points[i] = new BorderPoint( value1(i * stepSize ), value2( i * stepSize ), hulpTria) ;
        }
    } else {
     for (int i = 1; i < points.size()-1; ++i)
        {  points[i] = new BorderPoint( value1(1- i * stepSize ), value2(1-  i * stepSize ), hulpTria) ;
        }
    }
  //we left the last point out, as not to make rounding errors, since i * stepsize should be 1
  if (!reverse)
     { points[points.size()-1] = new BorderPoint( value1(1.0), value2(1.0), hulpTria);
     } else {
       points[points.size()-1] = new BorderPoint( value1(0.0), value2(0.0), hulpTria);
     }
}
double K2dBorderSinCurve::value1(double t) const
{ //only 3 digits eg 0.9998 must be 1, or 0.9994 must be 0.999
  int i  = (length * t + offset ) * 1000.0 + 0.5 ;
  double d = i / 1000.0 ;
  return d ;
}
double K2dBorderSinCurve::value2(double t) const
{
  int i;
  double h = -1.0 * ( (slope*t) + amplitude * sin(2.0 * M_PI * hills * t) ) *1000.0 ;
  //round to nearest double with 3 decimals
  if (h >= 0.0)
    { i = -1.0 * ( (slope*t) + amplitude * sin(2.0 * M_PI * hills * t) ) *1000.0 + 0.5;
    } else {
      i =  -1.0 * ( (slope*t) + amplitude * sin(2.0 * M_PI * hills * t) ) *1000.0 - 0.5;
    }
  double d = i/1000.0 ;
  return d ;
}
/**write a line to the ouputstream in a nice text format */
QTextStream &operator<<( QTextStream &os,  const K2dBorderSinCurve &sine )
{
  os << i18n("Sine Curve: a=") << sine.slope << " , V=" <<sine.amplitude << ", L=" << sine.length << ", offset = " << sine.offset ;
  os << ", n=" << sine.hills << i18n("\n     from (     ") ;
  os << sine.firstPoint()->coord1() << ", " <<   sine.firstPoint()->coord2() << i18n(" ) to (")      ;
  os << sine.lastPoint()->coord1() << ", " <<   sine.lastPoint()->coord2() ;
  os << i18n(") \n") ;
  return os;
}
