/***************************************************************************
                          k2ddomain.cpp  -  description
                             -------------------
    begin                : Mon Apr 30 2001
    copyright            : (C) 2001 by benny
    email                : bm@cage.rug.ac.be
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "k2ddomain.h"

/** Construction of a domain. Is a list with autodelete TRUE, since a border should be only added
     once, and has no meaning outside the domain */
K2dDomain::K2dDomain(){
   // if a border is removed from the domain, she may be deleted
   domain.setAutoDelete( TRUE );
}
K2dDomain::~K2dDomain(){
}
bool K2dDomain::closed()
{
  int i = 0;
  int count=domain.count();
  for (i = 0;  i <= count-1  ;  ++i)
     { if (i == 0)
         { if (  *( domain.at(count-1)->lastPoint() )   != *( domain.at(i)->firstPoint() )   )
                  { return false; }
         } else
         {
              if (  *( domain.at(i-1)->lastPoint() )  != *( domain.at(i)->firstPoint() )  )
                  { return false; }
         }
     }
  return true;
}
/** draw on a painter the domain (= the borders) */
void K2dDomain::draw(QPainter* p , bool meshed)
{
  int i = 0;
  int count=domain.count();
  for (i = 0;  i <= count-1  ;  ++i)
     { //draw every border seperately
       domain.at(i)->draw(p , meshed);
     }
}
/** write the domain to a textstream as required by the EasyMesh programm */
void K2dDomain::EasyMeshFile(QTextStream &cout)
{
  cout << "#----------------#\n# EasyMesh File #\n#----------------#\n"  ;
  cout << "\n# POINTS #\n"        ;
  cout << nrOfPoints() << " # number of points that will be inputed next#\n" ;

  cout << "\n# Nodes of the outer boundary#\n" ;
  int i=0;
  int j=0;
  int number=0;
  //we know the domain must be closed, so we never count the last point of a border, since it will be the first point of the next border
  for ( ; i<=nrOfBorders()-1; ++i )
    {
      for (j=0 ;j <= domain.at(i)->nrPointsOnBorder() - 2; ++j)
        {
          cout << number<<": "<<domain.at(i)->atPoint(j)->coord1() ;
          cout << " " << domain.at(i)->atPoint(j)->coord2() ;
          cout << " " << domain.at(i)->atPoint(j)->triaSize() ;
          //marker, for now we take a new one for each entered border !
          cout << " " << i+1 ;
          ++number;
          cout << "\n";
        }
    }
  cout << "\n#Nodes for refinement in the inner part - Adaptive Domain# \n" ;
  //now we must add the points of the holes or the adaptive domains. Not implemented yet

  cout << "\n# SEGMENTS #\n" ;
  cout << nrOfSegments() << " # number of segments that will be inputed next#\n" ;
  cout << "\n# Segments of the outer boundary#\n" ;
  //the max number of segments is the number of points.
  int maxsegment=number-1;
  number=0;
  //we count the same way as the points, but now we output the number of the points in a segment.
  for ( i=0; i<=nrOfBorders()-1; ++i )
    {
      for (j=0 ;j<=domain.at(i)->nrPointsOnBorder()-2; ++j)
        {
          if (number == maxsegment)
            {
              cout << number <<": " << number << " " << 0 ;
            } else {
              cout << number<<": "<< number <<" " << number+1    ;
            }
          //marker, for now we take a new one for each entered border !
          cout << " " << i+1 ;
          ++number;
          cout << "\n";
        }
    }
  cout << "\n#Segments for refinement in the inner part - Adaptive Domain# \n" ;
  //now we must add the segments of the holes or the adaptive domains. Not implemented yet
}
/** Return a listinterator over the borders in the domain */
QListIterator<K2dBorderBase> K2dDomain::iterator()
{
  QListIterator<K2dBorderBase>   hulpIter(domain);
  return hulpIter;
}
double K2dDomain::minCo(uint coord)
{
  QListIterator<K2dBorderBase>   it(domain);
  double minco;
  //init min
   minco = it.current()->firstPoint()->at(coord);
   for (  ; it.current() ; ++it )
       {
         it.current()->minCo(coord, minco);
       }
   return minco;
}
double K2dDomain::maxCo(uint coord)
{
  QListIterator<K2dBorderBase>   it(domain);
  double maxco;
  //init min
   maxco = it.current()->firstPoint()->at(coord);
   for (  ; it.current() ; ++it )
       {
         it.current()->maxCo(coord, maxco);
       }
   return maxco;
}
/** Return the number of points we represent the borders with */
uint K2dDomain::nrOfPoints()
{
 int i=0;
 int total = 0 ;
 //we know the domain must be closed, so we never count the last point of a border, since it will be the first point of the next border
 for (i=0; i<=nrOfBorders()-1; ++i )
   {
     total = total + domain.at(i)->nrPointsOnBorder() -1 ;
   }
 //now we must add the points of the holes or the adaptive domains. Not implemented yet
 return total;
}
/** Return the number of segments we represent the borders with */
uint K2dDomain::nrOfSegments()
{
  //for the closed outer domain we have as many segments as we have points.
  //since inner domains are not yet implemented, we can return the number of Points
  return nrOfPoints();
}
/* not yet implemented
K2dDomain K2dDomain::copy() const
{
  K2dDomain hulpDomain;
  //hulpDomain.domain.setAutoDelete( FALSE );
  DomainType hulpBorderList;
  for (i=0; i<=domain.count-1 ; ++i)
    {
    }
  hulpDomain.domain =hulpBorderList;
  return hulpDomain;
}
*/
QTextStream &operator<<( QTextStream & os , K2dDomain const &  domainOut)
{
  QListIterator<K2dBorderBase> iterDom(domainOut.domain);   //iterator points to beginning
  //if their are borders in the domain, print them in pretty text
  if (iterDom.current() )
   {  for ( ; iterDom.current(); ++iterDom )                //run trough list until iterator points at null
        { //os << *(iterDom.current())  ;
          iterDom.current()->printBorder(os);
        }
    } else {
      os << "No Domain Set" ;
    }
  return os;
}


