/***************************************************************************
                          k2dgrflow.cpp  -  description
                             -------------------
    begin                : Mon Apr 23 11:45:13 CEST 2001
    copyright            : (C) 2001 by benny
    email                : bm@cage.rug.ac.be
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

// include files for QT
#include <qdir.h>
#include <qprinter.h>
#include <qpainter.h>
#include <qmessagebox.h>
#include <qlineedit.h>
#include <qtextstream.h>
#include <qfile.h>

// include files for KDE
#include <kiconloader.h>
#include <kfiledialog.h>
#include <kmenubar.h>
#include <klocale.h>
#include <kconfig.h>
#include <kstdaction.h>
#include <krun.h>

// application specific includes
#include "k2dgrflow.h"
#include "k2dgrflowview.h"
#include "k2dgrflowdoc.h"
#include "dlgborder.h"
#include "k2ddomain.h"

//solution includes - should become link library
#include "domainnet2d.h"
#include "galarimpgrflow.h"

#define ID_STATUS_MSG 1

K2dGrFlowApp::K2dGrFlowApp(QWidget* , const char* name):KMainWindow(0, name)
{
  config=kapp->config();
  //create space for the domain;
  domain = new K2dDomain();
  ///////////////////////////////////////////////////////////////////
  // call inits to invoke all other construction parts
  initStatusBar();
  initActions();
  initDocument();
  initView();
	
  readOptions();

  ///////////////////////////////////////////////////////////////////
  // disable actions at startup
  filePrint->setEnabled(false);
  editCut->setEnabled(false);
  editCopy->setEnabled(false);
  editPaste->setEnabled(false);
  borderAdaptive->setEnabled(false);
  meshViewElements->setEnabled(false);
  meshViewNodes->setEnabled(false);
  meshViewSides->setEnabled(false);
}

K2dGrFlowApp::~K2dGrFlowApp()
{

}

void K2dGrFlowApp::initActions()
{
  fileNewWindow = new KAction(i18n("New &Window"),  0, this, SLOT(slotFileNewWindow()), actionCollection(),"file_new_window");
  fileNew = KStdAction::openNew(this, SLOT(slotFileNew()), actionCollection());
  fileOpen = KStdAction::open(this, SLOT(slotFileOpen()), actionCollection());
  fileOpenRecent = KStdAction::openRecent(this, SLOT(slotFileOpenRecent(const KURL&)), actionCollection());
  fileSave = KStdAction::save(this, SLOT(slotFileSave()), actionCollection());
  fileSaveAs = KStdAction::saveAs(this, SLOT(slotFileSaveAs()), actionCollection());
  fileClose = KStdAction::close(this, SLOT(slotFileClose()), actionCollection());
  filePrint = KStdAction::print(this, SLOT(slotFilePrint()), actionCollection());
  fileQuit = KStdAction::quit(this, SLOT(slotFileQuit()), actionCollection());
  editCut = KStdAction::cut(this, SLOT(slotEditCut()), actionCollection());
  editCopy = KStdAction::copy(this, SLOT(slotEditCopy()), actionCollection());
  editPaste = KStdAction::paste(this, SLOT(slotEditPaste()), actionCollection());
  viewToolBar = KStdAction::showToolbar(this, SLOT(slotViewToolBar()), actionCollection());
  viewStatusBar = KStdAction::showStatusbar(this, SLOT(slotViewStatusBar()), actionCollection());

  //self defined actions
  QPixmap pixMesh = MainBarIcon("k2dmesh");
  QIconSet iconMesh(pixMesh);
  QPixmap pixBorder = MainBarIcon("k2dborder");
  QIconSet iconBorder(pixBorder);
  QPixmap pixModel = MainBarIcon("k2dmodel");
  QIconSet iconModel(pixModel);
   borderInput = new KAction(i18n("Input &Borders"), iconBorder, 0, this, SLOT(slotBorderInput()),actionCollection(),"border_input");
   borderAdaptive = new KAction(i18n("Input &Adaptive Borders"), 0 , 0, this, SLOT(slotBorderAdaptive()),actionCollection(),"border_adaptive");
   //no icon for the following action (a parameter less !)
   borderPlot = new KAction(i18n("&Plot Borders"), 0, this, SLOT(slotBorderPlot()),actionCollection(),"border_plot");
   borderPlotMeshBorder = new KAction(i18n("Plot Borders &Mesh"), 0, this, SLOT(slotBorderPlotMeshBorder()),actionCollection(),"border_plot_mesh_border");
   meshOptions = new KAction(i18n("&Options"), 0, this, SLOT(slotMeshOptions()),actionCollection(),"mesh_options");
   meshGenerate = new KAction(i18n("&Generate"), iconMesh, 0, this, SLOT(slotMeshGenerate()),actionCollection(),"mesh_generate");
   meshPlot = new KAction(i18n("&Plot"), 0, this, SLOT(slotMeshPlot()),actionCollection(),"mesh_plot");
   meshViewNodes = new KAction(i18n("View &Nodes"),  0, this, SLOT(slotMeshViewNodes()),actionCollection(),"mesh_view_nodes");
   meshViewElements = new KAction(i18n("View &Elements"), 0, this, SLOT(slotMeshViewElements()),actionCollection(),"mesh_view_elements");
   meshViewSides = new KAction(i18n("View &Sides"), 0, this, SLOT(slotMeshViewSides()),actionCollection(),"mesh_view_sides");
   meshShowMesh = new KAction(i18n("E&xecute ShowMesh"), 0, this, SLOT(slotMeshShowMesh()),actionCollection(),"mesh_show_mesh");
   femRunModel = new KAction(i18n("&Run Model"), iconModel, 0, this, SLOT(slotFEMRunModel()),actionCollection(),"fem_run_model");
   femViewResult = new KAction(i18n("&View Result"), iconModel , 0, this, SLOT(slotFEMViewResult()),actionCollection(),"fem_view_result");

  fileNewWindow->setStatusText(i18n("Opens a new application window"));
  fileNew->setStatusText(i18n("Creates a new document"));
  fileOpen->setStatusText(i18n("Opens an existing document"));
  fileOpenRecent->setStatusText(i18n("Opens a recently used file"));
  fileSave->setStatusText(i18n("Saves the actual document"));
  fileSaveAs->setStatusText(i18n("Saves the actual document as..."));
  fileClose->setStatusText(i18n("Closes the actual document"));
  filePrint ->setStatusText(i18n("Prints out the actual document"));
  fileQuit->setStatusText(i18n("Quits the application"));
  editCut->setStatusText(i18n("Cuts the selected section and puts it to the clipboard"));
  editCopy->setStatusText(i18n("Copies the selected section to the clipboard"));
  editPaste->setStatusText(i18n("Pastes the clipboard contents to actual position"));
  viewToolBar->setStatusText(i18n("Enables/disables the toolbar"));
  viewStatusBar->setStatusText(i18n("Enables/disables the statusbar"));
   borderInput->setStatusText(i18n("Opens the dialog to set the Border"));
   borderAdaptive->setStatusText(i18n("Opens the dialog to set an adaptive border where mesh has to be finer or less finer than surrounding"));
   borderPlot->setStatusText(i18n("Plot the Border"));
   borderPlotMeshBorder->setStatusText(i18n("Plot the Border of the Mesh, this is an approximation of the real border"));
   meshOptions->setStatusText(i18n("Input Options of the Mesh"));
   meshGenerate->setStatusText(i18n("Generate the Mesh"));
   meshViewNodes->setStatusText(i18n("View file with the nodes"));
   meshViewElements->setStatusText(i18n("View file with the elements"));
   meshViewSides->setStatusText(i18n("View file with the sides of the elements"));
   meshShowMesh->setStatusText(i18n("Start the external program ShowMesh"));
   femRunModel->setStatusText(i18n("Run the model"));
   femViewResult->setStatusText(i18n("Plot the result of the model"));

  // use the absolute path to your k2dgrflowui.rc file for testing purpose in createGUI();
  createGUI();

}


void K2dGrFlowApp::initStatusBar()
{
  ///////////////////////////////////////////////////////////////////
  // STATUSBAR
  // TODO: add your own items you need for displaying current application status.
  statusBar()->insertItem(i18n("Ready."), ID_STATUS_MSG);
}

void K2dGrFlowApp::initDocument()
{
  doc = new K2dGrFlowDoc(this);
  doc->newDocument();
}

void K2dGrFlowApp::initView()
{ 
  ////////////////////////////////////////////////////////////////////
  // create the main widget here that is managed by KTMainWindow's view-region and
  // connect the widget to your document to display document contents.

  //extend the toolbar the toolbar with selfdefined possibilities
  borderInput->plug(toolBar());                                 // plug this action in the toolbar
  meshGenerate->plug(toolBar());                                 // plug this action in the toolbar
  femRunModel->plug(toolBar());                                 // plug this action in the toolbar

  view = new K2dGrFlowView(this);
  doc->addView(view);
  setCentralWidget(view);	
  setCaption(doc->URL().fileName(),false);
  show();
}

void K2dGrFlowApp::openDocumentFile(const KURL& url)
{
  slotStatusMsg(i18n("Opening file..."));
  //open the url and load the document
  bool ok = doc->openDocument( url);
  //if a document is found, load the domain
  if (ok)
    { //we init the domain
      K2dDomain *hulp;
      hulp = domain;
      domain = doc->get_domain() ;
      delete hulp;
      //add the url to the recent list
      fileOpenRecent->addURL( url );
    } else {
      //no document opened, also domain has to be empty
      K2dDomain *hulp;
      hulp = domain;
      domain = new K2dDomain();
      delete hulp;
    }
  //the view was inited in the beginning. Now we
  //load the stuff found in the document
  view->loadMyView();
  slotStatusMsg(i18n("Ready."));

}


K2dGrFlowDoc *K2dGrFlowApp::getDocument() const
{
  return doc;
}

K2dDomain *K2dGrFlowApp::getDomain() const
{
  return domain;
}

void K2dGrFlowApp::saveOptions()
{	
  config->setGroup("General Options");
  config->writeEntry("Geometry", size());
  config->writeEntry("Show Toolbar", viewToolBar->isChecked());
  config->writeEntry("Show Statusbar",viewStatusBar->isChecked());
  config->writeEntry("ToolBarPos", (int) toolBar("mainToolBar")->barPos());
  fileOpenRecent->saveEntries(config,"Recent Files");
}


void K2dGrFlowApp::readOptions()
{
	
  config->setGroup("General Options");

  // bar status settings
  bool bViewToolbar = config->readBoolEntry("Show Toolbar", true);
  viewToolBar->setChecked(bViewToolbar);
  slotViewToolBar();

  bool bViewStatusbar = config->readBoolEntry("Show Statusbar", true);
  viewStatusBar->setChecked(bViewStatusbar);
  slotViewStatusBar();


  // bar position settings
  KToolBar::BarPosition toolBarPos;
  toolBarPos=(KToolBar::BarPosition) config->readNumEntry("ToolBarPos", KToolBar::Top);
  toolBar("mainToolBar")->setBarPos(toolBarPos);
	
  // initialize the recent file list
  fileOpenRecent->loadEntries(config,"Recent Files");

  QSize size=config->readSizeEntry("Geometry");
  if(!size.isEmpty())
  {
    resize(size);
  }
}

void K2dGrFlowApp::saveProperties(KConfig *_cfg)
{
  if(doc->URL().fileName()!=i18n("Untitled") && !doc->isModified())
  {
    // saving to tempfile not necessary

  }
  else
  {
    KURL url=doc->URL();	
    _cfg->writeEntry("filename", url.url());
    _cfg->writeEntry("modified", doc->isModified());
    QString tempname = kapp->tempSaveName(url.url());
    QString tempurl= KURL::encode_string(tempname);
    KURL _url(tempurl);
    doc->saveDocument(_url);
  }
}


void K2dGrFlowApp::readProperties(KConfig* _cfg)
{
  QString filename = _cfg->readEntry("filename", "");
  KURL url(filename);
  bool modified = _cfg->readBoolEntry("modified", false);
  if(modified)
  {
    bool canRecover;
    QString tempname = kapp->checkRecoverFile(filename, canRecover);
    KURL _url(tempname);
  	
    if(canRecover)
    {
      doc->openDocument(_url);
      doc->setModified();
      setCaption(_url.fileName(),true);
      QFile::remove(tempname);
    }
  }
  else
  {
    if(!filename.isEmpty())
    {
      doc->openDocument(url);
      setCaption(url.fileName(),false);
    }
  }
}		

bool K2dGrFlowApp::queryClose()
{
  return doc->saveModified();
}

bool K2dGrFlowApp::queryExit()
{
  saveOptions();
  return true;
}

/////////////////////////////////////////////////////////////////////
// SLOT IMPLEMENTATION
/////////////////////////////////////////////////////////////////////

void K2dGrFlowApp::slotFileNewWindow()
{
  slotStatusMsg(i18n("Opening a new application window..."));
	
  K2dGrFlowApp *new_window= new K2dGrFlowApp();
  new_window->show();

  slotStatusMsg(i18n("Ready."));
}

void K2dGrFlowApp::slotFileNew()
{
  slotStatusMsg(i18n("Creating new document..."));

  if(!doc->saveModified())
  {
     // here saving wasn't successful

  }
  else
  {	
    doc->newDocument();
    //we init the domain
    K2dDomain *hulp;
    hulp = domain;
    domain = new K2dDomain();
    delete hulp;
    //we reload the view, since their is no document, this will create an empty view
    view->loadMyView();		

    setCaption(doc->URL().fileName(), false);
  }

  slotStatusMsg(i18n("Ready."));
}

void K2dGrFlowApp::slotFileOpen()
{
  slotStatusMsg(i18n("Opening file..."));
	
  if(!doc->saveModified())
  {
     // here saving wasn't successful

  }
  else
  {	
    KURL url=KFileDialog::getOpenURL(QString::null,
        i18n("*.gfm|Model files"), this, i18n("Open File..."));
    if(!url.isEmpty())
    {
      openDocumentFile(url);
      //doc->openDocument(url);
      //view->loadMyView();
      setCaption(url.fileName(), false);
      //fileOpenRecent->addURL( url );
    }
  }
  slotStatusMsg(i18n("Ready."));
}

void K2dGrFlowApp::slotFileOpenRecent(const KURL& url)
{
  slotStatusMsg(i18n("Opening file..."));
	
  if(!doc->saveModified())
  {
     // here saving wasn't successful
  }
  else
  {
    openDocumentFile(url);
    //doc->openDocument(url);
    //view->loadMyView();
    setCaption(url.fileName(), false);
  }

  slotStatusMsg(i18n("Ready."));
}

void K2dGrFlowApp::slotFileSave()
{
  //validation code start
  if (!checkModel())           //check what's available now
    {  slotStatusMsg(i18n("Unable to Save"));
       return;
  //validation code end
    } else
      //connect view with the doc as to update the doc. Like this, saving the doc will save what's on the screen
    { doc->put_d(view->led_dParam->text().toDouble() );
      doc->put_domain(domain);
    }

  slotStatusMsg(i18n("Saving file..."));
  if (doc->isNew()){
       slotFileSaveAs()   ;
    }else {
       doc->saveDocument(doc->URL());
       slotStatusMsg(i18n("Ready."));
  }
}

void K2dGrFlowApp::slotFileSaveAs()
{
  //validation code start
  if (!checkModel())
    {  slotStatusMsg(i18n("Unable to Save"));
       return;
  //validation code end
    } else
      //temperary storage of values before actual saving
      //connect view with the doc as to update the doc. Like this, saving the doc will save what's on the screen
    { doc->put_d(view->led_dParam->text().toDouble() );
      doc->put_domain(domain);
    }

  slotStatusMsg(i18n("Saving file with a new filename..."));

  KURL url=KFileDialog::getSaveURL(QDir::currentDirPath(),
        i18n(" *.gfm |Model files"), this, i18n("Save Model File as..."));
  if(!url.isEmpty())
  {
    //we add .gfm to the given filename
    QString filenm = url.fileName() + ".gfm";
    url.setFileName(filenm);
    doc->setURL(url);  //we set the URL of our document correct
    doc->saveDocument(url);
    fileOpenRecent->addURL(url);
    setCaption(url.fileName(),doc->isModified());
  }

  slotStatusMsg(i18n("Ready."));
}

void K2dGrFlowApp::slotFileClose()
{
  slotStatusMsg(i18n("Closing file..."));
	
  close();

  slotStatusMsg(i18n("Ready."));
}

void K2dGrFlowApp::slotFilePrint()
{
  slotStatusMsg(i18n("Printing..."));

  QPrinter printer;
  if (printer.setup(this))
  {
    view->print(&printer);
  }

  slotStatusMsg(i18n("Ready."));
}

void K2dGrFlowApp::slotFileQuit()
{
  slotStatusMsg(i18n("Exiting..."));
  saveOptions();
  // close the first window, the list makes the next one the first again.
  // This ensures that queryClose() is called on each window to ask for closing
  KMainWindow* w;
  if(memberList)
  {
    for(w=memberList->first(); w!=0; w=memberList->first())
    {
      // only close the window if the closeEvent is accepted. If the user presses Cancel on the saveModified() dialog,
      // the window and the application stay open.
      if(!w->close())
	break;
    }
  }	
  slotStatusMsg(i18n("Ready."));
}

void K2dGrFlowApp::slotEditCut()
{
  slotStatusMsg(i18n("Cutting selection..."));

  slotStatusMsg(i18n("Ready."));
}

void K2dGrFlowApp::slotEditCopy()
{
  slotStatusMsg(i18n("Copying selection to clipboard..."));

  slotStatusMsg(i18n("Ready."));
}

void K2dGrFlowApp::slotEditPaste()
{
  slotStatusMsg(i18n("Inserting clipboard contents..."));

  slotStatusMsg(i18n("Ready."));
}

void K2dGrFlowApp::slotViewToolBar()
{
  slotStatusMsg(i18n("Toggling toolbar..."));
  ///////////////////////////////////////////////////////////////////
  // turn Toolbar on or off
  if(!viewToolBar->isChecked())
  {
    toolBar("mainToolBar")->hide();
  }
  else
  {
    toolBar("mainToolBar")->show();
  }		

  slotStatusMsg(i18n("Ready."));
}

void K2dGrFlowApp::slotViewStatusBar()
{
  slotStatusMsg(i18n("Toggle the statusbar..."));
  ///////////////////////////////////////////////////////////////////
  //turn Statusbar on or off
  if(!viewStatusBar->isChecked())
  {
    statusBar()->hide();
  }
  else
  {
    statusBar()->show();
  }

  slotStatusMsg(i18n("Ready."));
}


void K2dGrFlowApp::slotStatusMsg(const QString &text)
{
  ///////////////////////////////////////////////////////////////////
  // change status message permanently
  statusBar()->clear();
  statusBar()->changeItem(text, ID_STATUS_MSG);
}

  /** Does the overal checking of the model before data is used or saved
       Generates Messageboxes with the errors, returns false if not ok, true when ok */
bool K2dGrFlowApp::checkModel()
{
  //first check if the d parameter is ok
  if (!view->validate_dParam())
    {QMessageBox::information( this, "K2dGrFlow",
                            "d parameter is wrongly set. It should be min 0.0, max 0.999999. \n"
                            "Change the value and try again" );

      return false;
    } ;
  return true;
}
/** slot that is called everytime the borderinput dialog is asked for */
void K2dGrFlowApp::slotBorderInput(){
   //call the dialog
  dlgBorder dlgbord(domain , this );
  int result = dlgbord.exec();
  //we reload the view so to see possible changes
  view->loadMyView();		


}
void K2dGrFlowApp::slotBorderAdaptive(){
}
void K2dGrFlowApp::slotBorderPlot(){
}
void K2dGrFlowApp::slotBorderPlotMeshBorder(){
}
void K2dGrFlowApp::slotMeshOptions(){
}
/** very big functionality that generates the mesh */
void K2dGrFlowApp::slotMeshGenerate()
{
  // !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  //first we validate if mesh generation is possible  : the model must be ok, it must have at least 3 borders and must be closed
  // !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  if (doc->isNew())
    {
     QMessageBox::information( this, "K2dGrFlow",
                            "Save the model first as a .gfm file. \n"
                            "Generate afterwards." );
     slotStatusMsg(i18n("Unable to Generate Mesh"));
     return;
    }
  if (!checkModel())
    {  slotStatusMsg(i18n("Unable to Generate Mesh"));
       return;
  //validation code end
    }
  if ( domain->nrOfBorders() < 3 )
    {
     QMessageBox::information( this, "K2dGrFlow",
                            "The domain has less than 3 borders. Generation is not possible. \n"
                            "Make sufficient borders, please." );
     slotStatusMsg(i18n("Unable to Generate Mesh"));
     return;
    }
  if ( !(domain->closed()) )
    {
     QMessageBox::information( this, "K2dGrFlow",
                            "The domain is not closed. Generation is not possible. \n"
                            "Make a closed domain and try again" );
     slotStatusMsg(i18n("Unable to Generate Mesh"));
     return;
    }
  //extra validations should be added, like : borders may not cross. This would lead us to far now.

  // !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  //now we write an outputfile that will be the input of our meshgenerating procedure
  // !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  QString basePath = QDir::currentDirPath() ;                    // eg /home/me/k2dgrflow/
  QString baseFileName =  doc->URL().fileName();            // test1.gfm
  baseFileName.remove(baseFileName.length()-4,4);          // remove the .gfm extension
  basePath.append("/" + baseFileName);                                     // eg /home/me/k2dgrflow/test1
  //the file we make will be a .d file. If such a file excists, ask if ok to overwrite it.
  bool exists;
  QFile *input= new QFile( basePath + ".d" );
  exists = input->exists();
  if (exists)
    {
     if (0==QMessageBox::information( this, "K2dGrFlow",
                            "There already exists a generated mesh.  \n"
                            "Ok to overwrite it with new generation?.","Cancel", "OK" ) )
        {
         slotStatusMsg(i18n("Unable to Generate Mesh"));
         return;       //we exit the mesh generation
        }
    }
  input->open(IO_WriteOnly);
 QTextStream cout( input);        // use a text stream
  domain->EasyMeshFile(cout);

  input->close();
  delete input;

  // !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  //now we execute the generation of the mesh
  // !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  pid_t  idEasyMesh = KRun::runCommand(QDir::currentDirPath() + "/./EasyMesh "+ baseFileName );
  //we let user follow the process
  QString idStr;
  idStr.sprintf("%i", idEasyMesh) ;
  QMessageBox::information( this, "K2dGrFlow",
                         "The meshing programm EasyMesh is started with pid "
                           + idStr +
                           "\n When finished you can continue. \n\n"
                          "In case of problem run manually /./EasyMesh "
                           + baseFileName +
                          ", and see error message."  );
  //we assume everything ok
  slotStatusMsg(i18n("Generation of the Mesh successfull ! "));

}
void K2dGrFlowApp::slotMeshPlot(){
}
void K2dGrFlowApp::slotMeshViewNodes(){
}
void K2dGrFlowApp::slotMeshViewElements(){
}
void K2dGrFlowApp::slotMeshViewSides(){
}
void K2dGrFlowApp::slotMeshShowMesh()
{
  //check if files are present to show the mesh
  QString basePath = QDir::currentDirPath() ;
  basePath.append("/"+doc->URL().fileName());          // eg /home/me/k2dgrflow/test1.gfm
  basePath.remove(basePath.length()-4,4);           // remove the .gfm extension
  bool ok;
  //do node file, element file and slide file exist ?
  QFile *fn = new QFile( basePath + ".n" );
  ok = fn->exists();
  delete fn;
  QFile *fe = new QFile( basePath + ".e" );
  ok = ok && fe->exists();
  delete fe;
  QFile *fs = new QFile( basePath + ".s" );
  ok = ok && fs->exists();
  delete fs;

  if (! ok)
    { //generate an error message
      QMessageBox::information( this, "K2dGrFlow",
                            "not possible to start ShowMesh. \n Necessary files do not exist. \n"
                            "Generate the mesh and try again" );
      //exit this action
      return ;
    }
  //everything ok, we run the command to start ShowMesh
  //notice that showmesh must be where the datafiles are. This is due to that program !
  QString command = QDir::currentDirPath()  + "/./ShowMesh "+ doc->URL().fileName();
  //the parameter to run showmesh is the file without extensions, so skip .gfm
  command.remove(command.length()-4,4) ;
  KRun::runCommand(command ) ;
}
void K2dGrFlowApp::slotFEMRunModel(){
  //check if files are present to show the mesh
  QString basePath = QDir::currentDirPath() ;
  basePath.append("/"+doc->URL().fileName());          // eg /home/me/k2dgrflow/test1.gfm
  basePath.remove(basePath.length()-4,4);           // remove the .gfm extension
  bool ok;
  //do node file, element file and slide file exist ?
  QFile *fn = new QFile( basePath + ".n" );
  ok = fn->exists();
  delete fn;
  QFile *fe = new QFile( basePath + ".e" );
  ok = ok && fe->exists();
  delete fe;
  QFile *fs = new QFile( basePath + ".s" );
  ok = ok && fs->exists();
  delete fs;

  if (! ok)
    { //generate an error message
      QMessageBox::information( this, "K2dGrFlow",
                            "not possible to run the model. \n Necessary files do not exist. \n"
                            "Generate the mesh and try again" );
      //exit this action
      return ;
    }
  //we first save the file, in case of a crash in the generation
  slotFileSave();
  //everything ok, we construct the problem and solve it, giving the progress to the user
  DomainNet2d *net = new DomainNet2d ;
  //now input the boundaries
  for (uint i=0 ; i < domain->nrOfBorders() ; ++i)
      {
        net->addBoundary(domain->border(i)) ;
      }
  KURL url ;
  url.setPath(basePath +".d");
  cout << i18n("Reading in the net from datafile...\n");
  net->buildNet(url);
  //now we construct the problem
  cout << i18n("Construct the problem to solve...\n");
  GalarImpGrFlow problem = GalarImpGrFlow(net , (doc->get_d()).toDouble());
  //solve the problem
  problem.solution();
  cout << "Saving solution to file <name>.u\n";
  QString foutputName(basePath + ".u");
  problem.printSolFile(foutputName.latin1());

  slotStatusMsg(i18n("Problem Solved! "));
  cout << i18n("Problem Solved! \n") ;
}
void K2dGrFlowApp::slotFEMViewResult(){
}
