/***************************************************************************
                          k2dgrflowdoc.cpp  -  description
                             -------------------
    begin                : Mon Apr 23 11:45:13 CEST 2001
    copyright            : (C) 2001 by benny
    email                : bm@cage.rug.ac.be
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

// include files for Qt
#include <qdir.h>
#include <qwidget.h>

// include files for KDE
#include <klocale.h>
#include <kmessagebox.h>
#include <kio/job.h>
#include <kio/netaccess.h>

// application specific includes
#include "k2dgrflowdoc.h"
#include "k2dgrflow.h"
#include "k2dgrflowview.h"
#include "k2ddomain.h"

QList<K2dGrFlowView> *K2dGrFlowDoc::pViewList = 0L;

K2dGrFlowDoc::K2dGrFlowDoc(QWidget *parent, const char *name) : QObject(parent, name)
{
  if(!pViewList)
  {
    pViewList = new QList<K2dGrFlowView>();
  }

  pViewList->setAutoDelete(true);

}

K2dGrFlowDoc::~K2dGrFlowDoc()
{
}

void K2dGrFlowDoc::addView(K2dGrFlowView *view)
{
  pViewList->append(view);
}

void K2dGrFlowDoc::removeView(K2dGrFlowView *view)
{
  pViewList->remove(view);
}
void K2dGrFlowDoc::setURL(const KURL &url)
{
  doc_url=url;
}

const KURL& K2dGrFlowDoc::URL() const
{
  return doc_url;
}

void K2dGrFlowDoc::slotUpdateAllViews(K2dGrFlowView *sender)
{
  K2dGrFlowView *w;
  if(pViewList)
  {
    for(w=pViewList->first(); w!=0; w=pViewList->next())
    {
      if(w!=sender)
        w->repaint();
    }
  }

}

bool K2dGrFlowDoc::saveModified()
{
  bool completed=true;

  if(modified)
  {
    K2dGrFlowApp *win=(K2dGrFlowApp *) parent();
    int want_save = KMessageBox::warningYesNoCancel(win, i18n("Warning"),
                                         i18n("The current file has been modified.\n"
                                              "Do you want to save it?"));
    switch(want_save)
    {
      case 1:
           if (doc_url.fileName() == i18n("Untitled"))
           {
             win->slotFileSaveAs();
           }
           else
           {
             saveDocument(URL());
       	   };

       	   deleteContents();
           completed=true;
           break;

      case 2:
           setModified(false);
           deleteContents();
           completed=true;
           break;	

      case 3:
           completed=false;
           break;

      default:
           completed=false;
           break;
    }
  }

  return completed;
}

void K2dGrFlowDoc::closeDocument()
{
  deleteContents();
}

bool K2dGrFlowDoc::newDocument()
{
  /////////////////////////////////////////////////
  // TODO: Add your document initialization code here
  modelFile.newFile();
  /////////////////////////////////////////////////
  modified=false;
  doc_url.setFileName(i18n("Untitled"));

  return true;
}

bool K2dGrFlowDoc::openDocument(const KURL& url, const char *format /*=0*/)
{
  QString tmpfile;
  KIO::NetAccess::download( url, tmpfile );
  /////////////////////////////////////////////////
  // TODO: Add your document opening code here
  //standard open file is performed at beginning of programm.
  // should do nothing if no url is specified
  bool open = false;
  if (!url.path().isEmpty() )
    {  modelFile.openFile(url);
       setURL(url);                    //set document URL so it points to new saving place
       open = true;                    //a document is found and opened
    }
  /////////////////////////////////////////////////

  KIO::NetAccess::removeTempFile( tmpfile );

  modified=false;
  return open;
}

bool K2dGrFlowDoc::saveDocument(const KURL& url, const char *format /*=0*/)
{
  /////////////////////////////////////////////////
  // TODO: Add your document saving code here
   modelFile.changeFileName(url.path());  //we make sure the path is correct
   modelFile.save();                                  //save code : delete old file, construct new out of memory
  /////////////////////////////////////////////////

  modified=false;
  return true;
}

void K2dGrFlowDoc::deleteContents()
{
  /////////////////////////////////////////////////
  // TODO: Add implementation to delete the document contents
  /////////////////////////////////////////////////

}

QString K2dGrFlowDoc::get_d()
{
   modelFile.setGroup("General Options");
   return modelFile.readEntry("dparam","0.0");
}
void K2dGrFlowDoc::put_d(const double d)
{
    //write to file. Will only be active on save
   modelFile.setGroup("General Options");
   modelFile.writeEntry("dparam", d );
}
   /** get the domain from disk  */
K2dDomain *K2dGrFlowDoc::get_domain()
{
   K2dDomain *hulpDom = new K2dDomain();
   int i = 1;
   QString iStr;
   K2dBorderLine* testLine;  //our temporary storage of the lines we will add to the domain
   iStr.setNum(i);
   while (modelFile.hasGroup("Border" + iStr) )
     { //there is a border entry, we start retrieving the info
       modelFile.setGroup("Border" + iStr);
       if  ( modelFile.readEntry("type") == "Line"  )
           {//we start to read in the data of a line, and add it a domain object
            // that we return in the end
            BorderPoint a(modelFile.readEntry("x1").toDouble(), modelFile.readEntry("z1").toDouble());
            BorderPoint b(modelFile.readEntry("x2").toDouble(), modelFile.readEntry("z2").toDouble());
            testLine = new K2dBorderLine( a , b);
            //now we add the trianglesized
            testLine->changeTriaSize(modelFile.readEntry("triangle1").toDouble(), K2dBorderLine::BeginPoint);
            testLine->changeTriaSize(modelFile.readEntry("triangle2").toDouble(), K2dBorderLine::EndPoint);
            //now we set the stuff off the base class, with the specific function written for that
            get_domainBase(testLine);
            //now we add the constructed sine to the domain
            hulpDom->appendBorder(testLine);
            ++i;
            iStr.setNum(i);
           }
       if  ( modelFile.readEntry("type") == "SinCurve"  )
         {
          K2dBorderSinCurve* hulpSine = new K2dBorderSinCurve( modelFile.readEntry("L").toDouble(),
                     modelFile.readEntry("a").toDouble(),  modelFile.readEntry("V").toDouble(), modelFile.readEntry("n").toUInt(),
                     modelFile.readEntry("b").toDouble() );
          //now we add the trianglesize
          hulpSine->changeTriaSize(modelFile.readEntry("triangle").toDouble());
          //set other values
          hulpSine->setReverse(modelFile.readBoolEntry("reverse")) ;
          hulpSine->setNrMeshInterval(modelFile.readEntry("meshpoints").toInt() );
          //now we set the stuff off the base class, with the specific function written for that
          get_domainBase(hulpSine);
          //now we add the constructed sine to the domain
          hulpDom->appendBorder(hulpSine);
          ++i;
          iStr.setNum(i);
         }
     } //end of while loop over the borders
   return hulpDom;
}
void K2dGrFlowDoc::get_domainBase(K2dBorderBase *border)
{
  //now we add bordercondtion to the border
  if (modelFile.hasKey("CondType" ) )
    { //first we set the condition to not set, then we put it to the actual value
      border->setBorderCond(K2dBorderBase::NotSet );
      if (modelFile.readEntry("CondType") == "NormalDer_0" )
        {
         border->setBorderCond(K2dBorderBase::NEUMANN);
        }
      if (modelFile.readEntry("CondType") == "Dirichlet" )
        {
         border->setBorderCond(K2dBorderBase::DIRICHLET);
         if (modelFile.readEntry("CondTypeDir") == "DirZ" )
           {
            border->setBorderCond(K2dBorderBase::DIRICHLET, K2dBorderBase::DirZ);
           }
         if (modelFile.readEntry("CondTypeDir") == "DirAx_B" )
           {
             border->setBorderCond(K2dBorderBase::DIRICHLET, K2dBorderBase::DirAx_B,
                     modelFile.readEntry("CondTypeDirLineA").toDouble(), modelFile.readEntry("CondTypeDirLineB").toDouble() );
           }
        }
    }
}
/** write the domain to disk  */
void K2dGrFlowDoc::put_domain(K2dDomain * dom)
{
  int i = 1;
  QString iStr;
  QListIterator<K2dBorderBase> iterDom =  dom->iterator();    //iterator points to beginning
  for ( ; iterDom.current(); ++iterDom )                //run trough list until iterator points at null
     { //write entries to our file
       iStr.setNum(i);
       //everything goes in the group that corresponds with the border
       modelFile.setGroup("Border" + iStr) ;
       ++i;
       if (iterDom.current()->type() == K2dBorderBase::Line )
         { //first we do a dynamic cast of a borderbase to a borderline
           K2dBorderLine* tempHelp = dynamic_cast <K2dBorderLine *> ( iterDom.current() );
           //enter the keys with their values
           modelFile.writeEntry("type", "Line");
           modelFile.writeEntry("x1" , iterDom.current()->firstPoint()->at(0) );
           modelFile.writeEntry("z1" , iterDom.current()->firstPoint()->at(1) );
           modelFile.writeEntry("x2" , iterDom.current()->lastPoint()->at(0) );
           modelFile.writeEntry("z2" , iterDom.current()->lastPoint()->at(1) );
           //now the given triangle sizes
           modelFile.writeEntry("triangle1" , tempHelp->triaSize(K2dBorderLine::BeginPoint) );
           modelFile.writeEntry("triangle2" , tempHelp->triaSize(K2dBorderLine::EndPoint) );
         }
       if (iterDom.current()->type() == K2dBorderBase::SinCurve )
         { //first we do a dynamic cast of a borderbase to a borderline
           K2dBorderSinCurve* tempHelp = dynamic_cast <K2dBorderSinCurve *> ( iterDom.current() );
           //enter the keys with their values
           modelFile.writeEntry("type", "SinCurve");
           modelFile.writeEntry("L" , tempHelp->getLength() );
           modelFile.writeEntry("a" , tempHelp->getSlope() );
           modelFile.writeEntry("V" ,tempHelp->getAmplitude() );
           modelFile.writeEntry("n" , tempHelp->getHills());
           modelFile.writeEntry("b" , tempHelp->getOffset());
           //now the given triangle size
           modelFile.writeEntry("triangle" , tempHelp->triaSize() );
           //other values
           modelFile.writeEntry("meshpoints" , tempHelp->nrMeshIntervalPoints() );
           modelFile.writeEntry("reverse" , tempHelp->reverseValue() );
         }
      //now we write away the bordercondition
      if (iterDom.current()->getBorderCondition() == K2dBorderBase::NotSet )
        {
          modelFile.writeEntry("CondType", "NotSet");
        }
      if (iterDom.current()->getBorderCondition() == K2dBorderBase::NEUMANN )
        { //the only neumann condition we have for now is normal derivative derivative zero
          //just like dirichlet this has to be changed to a condtype and a conttypeneu
          modelFile.writeEntry("CondType", "NormalDer_0");
        }
      if (iterDom.current()->getBorderCondition() == K2dBorderBase::DIRICHLET)
        {
          modelFile.writeEntry("CondType", "Dirichlet");
          if (iterDom.current()->getBorderCondDirType() == K2dBorderBase::DirZ)
            {
              modelFile.writeEntry("CondTypeDir","DirZ");
            }
          if (iterDom.current()->getBorderCondDirType() == K2dBorderBase::DirAx_B)
            {
              modelFile.writeEntry("CondTypeDir","DirAx_B");
              modelFile.writeEntry("CondTypeDirLineA", iterDom.current()->getBorderCondDirLineA() );
              modelFile.writeEntry("CondTypeDirLineB", iterDom.current()->getBorderCondDirLineB() );
            }
        }
     }  //end for loop over all the borders
}
