/***************************************************************************
                          k2dgrflowview.cpp  -  description
                             -------------------
    begin                : Mon Apr 23 11:45:13 CEST 2001
    copyright            : (C) 2001 by benny
    email                : bm@cage.rug.ac.be
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

// include files for Qt
#include <qprinter.h>
#include <qpainter.h>
#include <qlineedit.h>
#include <qvalidator.h>
#include <qmultilineedit.h>
#include <qpaintdevice.h>
#include <qtextstream.h>
#include <qlayout.h>
#include <qpushbutton.h>

#include <klocale.h>
#include <kapp.h>

// application specific includes
#include "grflowviewdesign.h"
#include "k2dgrflowview.h"
#include "k2dgrflowdoc.h"
#include "k2dgrflow.h"
#include "k2dborderbase.h"
#include "k2ddomain.h"
#include "plotfield.h"

K2dGrFlowView::K2dGrFlowView(QWidget *parent, const char *name) : GrFlowViewDesign(parent, name)
{
  /** We want a scrollable area on the right. This is not possible with the designer, so we add it manually
        by adding the class PlotField, our class that implements the drawing routines
        We also pass a pointer to the view since view is not parent, but a frame is. */
    pfResult = new PlotField(this , Frame4, "pfResult");
    //the plotfield is a part of Layout15 of our design, a Vertical layout box together with our command boxes
    Layout15->addWidget( pfResult );
    //now some extra buttons for under the plotfield :
    cmdZoomOut = new QPushButton( Frame4);
    cmdZoomOut->setEnabled( TRUE );
    cmdZoomOut->setText( i18n( "Zoom Out" ) );
    cmdZoomIn = new QPushButton( Frame4);
    cmdZoomIn->setEnabled( TRUE );
    cmdZoomIn->setText( i18n( "Zoom In" ) );
    cmdPrint = new QPushButton( Frame4);
    cmdPrint->setEnabled( TRUE );
    cmdPrint->setText( i18n( "Print" ) );

    layoutZoom = new QHBoxLayout;
    layoutZoom->setSpacing( 6 );
    layoutZoom->setMargin( 0 );
    layoutZoom->addWidget(cmdZoomOut);
    layoutZoom->addWidget(cmdZoomIn);
    layoutZoom->addWidget(cmdPrint);
    Layout15->addLayout(layoutZoom);

    //make extra connects between signals and slots
    connect( cmdDomain, SIGNAL(clicked()), pfResult, SLOT(plotDomain()) );
    connect( cmdBorderMesh, SIGNAL(clicked()), pfResult, SLOT(plotDomainMesh()) );
    connect( cmdMesh, SIGNAL(clicked()), pfResult, SLOT(plotMesh()) );
    connect( cmdResult, SIGNAL(clicked()), pfResult, SLOT(plotResult()) );
    connect( cmdZoomOut, SIGNAL(clicked()), pfResult, SLOT(zoomOut()) );
    connect( cmdZoomIn, SIGNAL(clicked()), pfResult, SLOT(zoomIn()) );
    connect( cmdPrint, SIGNAL(clicked()), pfResult, SLOT(print()) );

    K2dGrFlowApp *theApp=(K2dGrFlowApp *) parentWidget();
    //make sure plot is initialized when we open new documents
    connect( theApp->fileClose, SIGNAL(activated()), pfResult, SLOT(blank()) );
    connect( theApp->fileNew, SIGNAL(activated()), pfResult, SLOT(blank()) );
    connect( theApp->fileOpen, SIGNAL(activated()), pfResult, SLOT(blank()) );
    connect( theApp->fileOpenRecent, SIGNAL(activated()), pfResult, SLOT(blank()) );
    //connect the different menu options to action of new elements on the view
    connect( theApp->borderPlot, SIGNAL(activated()), pfResult, SLOT(plotDomain()) );
    connect( theApp->borderPlotMeshBorder, SIGNAL(activated()), pfResult, SLOT(plotDomainMesh()) );
    connect( theApp->meshPlot, SIGNAL(activated()), pfResult, SLOT(plotMesh()) );
    connect( theApp->filePrint, SIGNAL(activated()), pfResult, SLOT(print()) );

    // extend the user interface with options that cannot be designed automatically
    // for d only small positive doubles are allowed, therefore connect a validator to the input zone
    led_dParam->setValidator(new QDoubleValidator( 0.0, 0.99999, 6, led_dParam));
}

K2dGrFlowView::~K2dGrFlowView()
{
}

K2dGrFlowDoc *K2dGrFlowView::getDocument() const
{
  K2dGrFlowApp *theApp=(K2dGrFlowApp *) parentWidget();

  return theApp->getDocument();
}
K2dDomain *K2dGrFlowView::getDomain() const
{
  K2dGrFlowApp *theApp=(K2dGrFlowApp *) parentWidget();

  return theApp->getDomain();
}

void K2dGrFlowView::print(QPrinter *pPrinter)
{
  QPainter printpainter;
  printpainter.begin(pPrinter);
	
  // TODO: add your printing code here

  printpainter.end();
}
/** Validate the input of the d parameter. Return true when ok, false when not ok. 
It is left to the user to change the d parameter until ok */
bool K2dGrFlowView::validate_dParam(){
 int i=1 ;
 QString tekst =  led_dParam->text() ;
 if (led_dParam->validator()->validate( tekst , i)  == QValidator::Acceptable )
   {return true; }
 else
   {return false;}
}
/** wil restate the View. Call the function every time you want to reload the view from the current document */
void K2dGrFlowView::loadMyView(){
     K2dGrFlowDoc *tmpDoc = getDocument();
     //set the d parameter
     led_dParam->setText(tmpDoc->get_d());
    //display the domain in a multilineedit box in pretty text
    QString str;
    QTextStream ts( &str, IO_WriteOnly );
    K2dGrFlowApp *theApp=(K2dGrFlowApp *) parentWidget();
    ts<< *(theApp->domain) ;    //View is friend of the application and has access to private parts !
    mleDomainView->setText(str);
}
