/***************************************************************************
                          plotfield.cpp  -  description
                             -------------------
    begin                : Thu May 3 2001
    copyright            : (C) 2001 by benny
    email                : bm@cage.rug.ac.be
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
//to read in the result (to plot it), we need the library for big vectors
#define __BENNY_CLOG
extern "C" {
   #include "meschach/matrix.h"
   #include "meschach/matrix2.h"
   #include "meschach/iter.h"
}

#include "plotfield.h"
#include "qarray.h"
#include <qcolor.h>
#include <qdir.h>
#include <qmessagebox.h>
#include <qpalette.h>
#include <qpainter.h>
#include <qprinter.h>

#include <fstream.h>

#include <kapp.h>

#include "k2ddomain.h"
#include "k2dgrflowview.h"
#include "k2dgrflowdoc.h"
//include solver for method to read in a VEC
#include "galarsolver2d.h"

PlotField::PlotField(K2dGrFlowView * view, QWidget *parent, const char *name ):QScrollView(parent,name)
{
   //we construct the internal of the PlotField.
   plotWidget = new PlotFieldView(view, this->viewport());
   //and add it to the viewport
   this->addChild(plotWidget);
   //we center it on the bottom left, since coordinate basis is put there.
   this->center(0, PlotFieldView::fieldHeight ) ;
}
PlotField::~PlotField(){
}
void PlotField::blank()
{    //we just pass the command
     plotWidget->updateIt(PlotFieldView::Blank);
}
void PlotField::plotDomain()
{    //we just pass the command
     plotWidget->updateIt(PlotFieldView::Domain);
}
void PlotField::plotDomainMesh()
{    //we just pass the command
     plotWidget->updateIt(PlotFieldView::DomainMesh);
}
void PlotField::plotMesh()
{ //checking if result excists
  QString basePath = QDir::currentDirPath() ;
  basePath.append("/"+ plotWidget->getOwner()->getDocument()->URL().fileName());          // eg /home/me/k2dgrflow/test1.gfm
  basePath.remove(basePath.length()-4,4);           // remove the .gfm extension
  bool ok;
  //does side file exist ?
  QFile *f = new QFile( basePath + ".s" );
  ok = f->exists();
  delete f;
  if (! ok)
    { //generate an error message
      QMessageBox::information( this, "K2dGrFlow",
                            "not possible to plot mesh. \n Necessary files do not exist. \n"
                            "Generate the mesh and try again" );
      //exit this action
      return ;
    }
  //everything ok
  //we pass the command to show the mesh
  plotWidget->updateIt(PlotFieldView::Mesh);
}
void PlotField::plotResult()
{  //checking if result excists
   QString basePath = QDir::currentDirPath() ;
   basePath.append("/"+ plotWidget->getOwner()->getDocument()->URL().fileName());          // eg /home/me/k2dgrflow/test1.gfm
   basePath.remove(basePath.length()-4,4);           // remove the .gfm extension
   bool ok;
   //does file with the unknowns, nodes, sides and elements exist ?
   QFile *f = new QFile( basePath + ".n" );
   ok = f->exists();
   delete f;
   f = new QFile( basePath + ".s" );
   ok = ok && f->exists();
   delete f;
   f = new QFile( basePath + ".e" );
   ok = ok &&f->exists();
   delete f;
   if (! ok)
     { //generate an error message
       QMessageBox::information( this, "K2dGrFlow",
                            "not possible to plot result. \n Mesh does not exist yet. \n"
                            "Generate the mesh, solve the problem, and try again" );
       //exit this action
       return ;
     }
   //see if solution excsts
   f = new QFile( basePath + ".u" );
   ok = f->exists();
   delete f;
   if (! ok)
     { //generate an error message
       QMessageBox::information( this, "K2dGrFlow",
                            "not possible to plot result. \n Solution not generated yet. \n"
                            "Solve the problem, and try again" );
       //exit this action
       return ;
     }
   //everything ok, we just pass the command
   plotWidget->updateIt(PlotFieldView::Result);
}
void PlotField::print()
{    //we just pass the command
     plotWidget->print();
}
void PlotField::printDomain()
{    //we just pass the command
     plotWidget->printIt(PlotFieldView::Domain);
}
void PlotField::zoomOut()
{    //we just pass the command
     plotWidget->zoomOut();
}
void PlotField::zoomIn()
{    //we just pass the command
     plotWidget->zoomIn();
}


PlotFieldView::PlotFieldView( K2dGrFlowView* owner, QWidget *parent, const char *name ):QWidget(parent,name)
{
    drawViewOwner = owner;
    setBackgroundColor( white );
    //we make the widget of size 1500 pixels wide and 2000 high
    QSize s(fieldWidth,fieldHeight);
    resize( s );
    //standard nothing is drawn
    drawindex = Blank;
    scaleHor = 0;
    scaleVer = 0;
    transHor = 0;
    transVer =0 ;
    zoomFactor = 1.0;
    printer = new QPrinter;
}

/** Clean up */
PlotFieldView::~PlotFieldView()
{
#ifndef QT_NO_PRINTER
    delete printer;
#endif
}
/**  Called whenever a certain contents must be drawn */
void PlotFieldView::updateIt(DrawContents dw)
{
   drawindex = dw;   //set index so we now what must be drawn
   //recalculate optimal plotting parameters
   setPlotValues();
   update();              //update will invoke a paintevent that does the actual drawing
}
/**  Calls the drawing function as specified by the drawindex. */

void PlotFieldView::drawIt( QPainter *p )
{
    p->save();
    //we wish a mathematical coordinate system (right - handed)
    p->scale(1,-1);
    //at the bottom left corner.
    p->translate(0,-1 * fieldHeight);
    //now use scale and translation necessary to show a nice figure
    p->translate(transHor * scaleHor * zoomFactor, transVer * scaleVer * zoomFactor );
    p->scale(scaleHor * zoomFactor  , scaleVer * zoomFactor);
   //we leave a border of 10 pixels
   p->translate (10.0 /(zoomFactor * scaleHor) ,10.0 / (zoomFactor * scaleVer));

    if (drawindex == Domain )
      {
         drawViewOwner->getDomain()->draw(p, false );
      };
    if (drawindex == DomainMesh )
      {
         drawViewOwner->getDomain()->draw(p, true );
      };
    int sides, nodes, ele ;
   QString dummy ;
   QString basePath = QDir::currentDirPath() ;
    basePath.append("/"+ getOwner()->getDocument()->URL().fileName());          // eg /home/me/k2dgrflow/test1.gfm
    basePath.remove(basePath.length()-4,4);           // remove the .gfm extension
    if (drawindex == Mesh )
      { // put a textstream on the file to read input
       QFile *fs = new QFile( basePath + ".s" );
       fs->open( IO_ReadOnly );
       QTextStream t( fs );
       t >> sides ;
       QArray<sid> sidesMesh(sides);
       int n1, n2, e1, e2, m ;
       int s;
       for(s=0; s<sides; s++)
        {
         t >> dummy >> n1 >> n2 >> e1 >> e2 >> m ;
         sidesMesh[s].node1 = n1;
         sidesMesh[s].node2 = n2;
        }
       fs->close();
       delete fs;
       QFile *fn = new QFile( basePath + ".n" );
       fn->open( IO_ReadOnly );
       QTextStream tn( fn );
       tn >> nodes ;

       QArray<nod> nodesMesh(nodes);
       for(int n=0; n<nodes; n++)
        {
         tn >> dummy >> nodesMesh[n].coord1 >> nodesMesh[n].coord2 >> m ;
        }
       fn->close();
       delete fn;
       //plot the mesh
       //we have to print our mesh ultimately as integer points on the sceen (pixels). Carefull handling is necessary
       //to be certain of a nice resolution we blow our mesh up and scale to a much smaller version
       p-> scale (1.0/ 1000.0, 1.0/1000.0);
       p->setPen(Qt::black);
       for (s=0; s<sides ; s++)
        {
          p->drawLine(nodesMesh[sidesMesh[s].node1].coord1* 1000.0,
                       nodesMesh[sidesMesh[s].node1].coord2 * 1000.0 ,
                       nodesMesh[sidesMesh[s].node2].coord1 * 1000.0 ,
                       nodesMesh[sidesMesh[s].node2].coord2 *1000.0);
        }
      };
    //plot the result as contours (only this implemented now)
    if (drawindex == Result )
      { //first read in the unknowns out of file.u in VEC format
        FILE *finputFile;
        QString finputName(basePath + ".u");
        finputFile = fopen( finputName.latin1() , "r");
        VEC *result = GalarSolver2d::readVEC(finputFile, VNULL);
        fclose (finputFile);
        //change this later to a double *contours wich has variable length
        double contours[nrCont+1];
        QColor contColor[nrCont+1];
        //min and max value in result;
        double minCont = result->ve[0];
        double maxCont = result->ve[0];
        int c, e, n1,n2,n3, e1,e2,e3,s1,s2,s3, m;
        double  xv,yv;
        for (c=1 ; c < result->dim; ++c)
            { if (result->ve[c] < minCont){minCont = result->ve[c]; }
              else if (result->ve[c] > maxCont){maxCont = result->ve[c]; }
            }
        if (maxCont == minCont)
          { cerr << "constant solution, can not plot contourplot\n" ;
          } else {
            //we decide the 20 contourvalues, and the 20 colours
            double diff = (maxCont - minCont) / (nrCont +1);
            int diffColor = 255 / (nrCont+1) ;
            for (c=0; c<= nrCont ; ++c)
                 { contours[c] = minCont + c * diff ;
                   contColor[c] = QColor(255 - c*diffColor, 0 , 0 + c*diffColor);
                 }
           //we now have 20 countours between the min and max value
           //per element, we plot. Since elements can be small, set everything to plot small things
           //to be certain of a nice resolution we blow our mesh up and scale to a much smaller version
           p-> scale (1.0/ 1000.0, 1.0/1000.0);
           //read in the nodes
           QFile *fn = new QFile( basePath + ".n" );
           fn->open( IO_ReadOnly );
           QTextStream tn( fn );
           tn >> nodes ;

           QArray<nod> nodesMesh(nodes);
           for(int n=0; n<nodes; n++)
            {
              tn >> dummy >> nodesMesh[n].coord1 >> nodesMesh[n].coord2 >> m ;
            }
           fn->close();
           delete fn;
           QFile *fe = new QFile( basePath + ".e" );
           fe->open( IO_ReadOnly );
           QTextStream t( fe );
           t >> ele ;
           //result has the result per node. If we now read in the elements, 1 per one, we can draw contours in the elements
           vector<int> vectCont12;
           vector<int> vectCont23;
           vector<int> vectCont31;
           int contNr, search;
           double xbCont,ybCont,xeCont,yeCont;
           for(e=0; e<ele; e++)
            {
              t >> dummy >> n1 >> n2 >> n3>>e1 >> e2 >>e3 >>s1 >> s2 >> s3 >> xv >> yv >> m ;
              contoursBetween(result->ve[n1], result->ve[n2], contours, vectCont12);
              contoursBetween(result->ve[n3], result->ve[n1], contours, vectCont31);
              contoursBetween(result->ve[n2], result->ve[n3], contours, vectCont23);
              while (vectCont12.size() !=0)
                { contNr = vectCont12.back();
                  vectCont12.pop_back();
                  //set the color
                  p->setPen(contColor[contNr]);
                  //determen coordinates on line 12 of the contour
                  if (!posCont(contours[contNr], nodesMesh[n1].coord1, nodesMesh[n1]. coord2 , result->ve[n1],
                                         nodesMesh[n2].coord1, nodesMesh[n2]. coord2,  result->ve[n2], xbCont , ybCont) )
                    { //the line itself is the contour, no beginpoint, draw that line as the contour
                      p->drawLine(nodesMesh[n1].coord1* 1000.0,  nodesMesh[n1].coord2 * 1000.0 ,
                                                nodesMesh[n2].coord1 * 1000.0 , nodesMesh[n2].coord2 *1000.0);
                    } else {
                      //search the contour on the other lines, first on 23
                      for (search=vectCont23.size()-1; search >-1 && vectCont23[search] != contNr; --search){};
                      if (!(search == -1))
                        { //found the countour on line 23, erase it there
                          vectCont23[search]=-1;
                          //determine position on line 23
                          posCont(contours[contNr], nodesMesh[n2].coord1, nodesMesh[n2]. coord2 , result->ve[n2],
                                            nodesMesh[n3].coord1, nodesMesh[n3]. coord2,  result->ve[n3], xeCont , yeCont);
                          //draw the contour
                          p->drawLine(xbCont * 1000.0, ybCont * 1000.0 , xeCont  * 1000.0 , yeCont * 1000.0);
                        } else {
                          //the contour is not on 23, it must be on 31
                          for (search=vectCont31.size()-1; vectCont31[search] != contNr; --search){}
                          //now we have the countour on 31, erase out of list, and determine position
                          vectCont31[search]=-1;
                          posCont(contours[contNr], nodesMesh[n3].coord1, nodesMesh[n3]. coord2 , result->ve[n3],
                                            nodesMesh[n1].coord1, nodesMesh[n1]. coord2, result->ve[n1], xeCont , yeCont);
                          //draw the contour
                          p->drawLine(xbCont * 1000.0, ybCont * 1000.0 , xeCont  * 1000.0 , yeCont * 1000.0);
                        }
                    }
                }
              //now 12 is finished, see what has to be done for 23
              while (vectCont23.size() !=0)
                { contNr = vectCont23.back() ;
                  vectCont23.pop_back();
                  if (contNr != -1)
                  {
                  //set the color
                  p->setPen(contColor[contNr]);
                  //determen coordinates on line 23 of the contour
                  if (!posCont(contours[contNr], nodesMesh[n2].coord1, nodesMesh[n2]. coord2 , result->ve[n2],
                                         nodesMesh[n3].coord1, nodesMesh[n3]. coord2, result->ve[n3], xbCont , ybCont) )
                    { //the line itself is the contour, no beginpoint, draw that line as the contour
                      p->drawLine(nodesMesh[n2].coord1* 1000.0,
                         nodesMesh[n2].coord2 * 1000.0 ,
                         nodesMesh[n3].coord1 * 1000.0 ,
                         nodesMesh[n3].coord2 *1000.0);
                    } else {
                      //the contour is not on 23, it must be on 31
                      for (search=vectCont31.size()-1; vectCont31[search] != contNr; --search){}
                      //now we have the countour on 31, erase out of list, and determine position
                      vectCont31[search]=-1;
                      posCont(contours[contNr], nodesMesh[n3].coord1, nodesMesh[n3]. coord2 , result->ve[n3],
                                         nodesMesh[n1].coord1, nodesMesh[n1]. coord2, result->ve[n1], xeCont , yeCont);
                      //draw the contour
                      p->drawLine(xbCont * 1000.0, ybCont * 1000.0 , xeCont  * 1000.0 , yeCont * 1000.0);
                    }
                  }
                }
              //now 23 is finished, it is still possible that on 31 there is a contour, albeit only the line itself !
              while (vectCont31.size() !=0)
                { contNr = vectCont31.back() ;
                  vectCont31.pop_back();
                  if (contNr !=-1)
                  {
                  //set the color
                  p->setPen(contColor[contNr]);
                  //determen coordinates on line 23 of the contour
                  if (!posCont(contours[contNr], nodesMesh[n2].coord1, nodesMesh[n2]. coord2 , result->ve[n2],
                                         nodesMesh[n3].coord1, nodesMesh[n3]. coord2, result->ve[n3], xbCont , ybCont) )
                    { //the line itself is the contour, no beginpoint, draw that line as the contour
                      p->drawLine(nodesMesh[n3].coord1* 1000.0,
                         nodesMesh[n3].coord2 * 1000.0 ,
                         nodesMesh[n1].coord1 * 1000.0 ,
                         nodesMesh[n1].coord2 *1000.0);
                    } else {
                      //search the contour on the other line is not possible (done already), so something is wrong
                      cerr << "procedure with contours contains error..\n";
                    }
                  }
                }
            }
           fe->close();
           delete fe;
           //now we draw the lines around the contour !
           p-> scale (1000.0, 1000.0);
           p->setPen(Qt::black);
           drawViewOwner->getDomain()->draw(p, false );
          }
      };
    p->restore();  //restore original coordinate set, in case someone still uses the painter in the paintevent
}

/**  Called when the print button is clicked, print what's in the printing area. */
void PlotFieldView::print()
{
   printIt(drawindex);
}
/** Printing a certain contents, even when it's not shown now */
void PlotFieldView::printIt(DrawContents dw)
{
   //we set the drawindex to what has to be printed
   DrawContents hulpDw = drawindex;
   drawindex = dw;
#ifndef QT_NO_PRINTER
   if ( printer->setup( this ) ) {
        QPainter paint( printer );
        drawIt( &paint );
        }
#endif
   //we set drawindex back to the original value, since that is the value that's still on screen
   drawindex= hulpDw;
}
/** Called when the widget needs to be updated. */
void PlotFieldView::paintEvent( QPaintEvent * )
{
    QPainter paint( this );
    drawIt( &paint );
}

/**private function that sets optimal drawing values */
void PlotFieldView::setPlotValues()
{
   if (drawindex == Blank)
     { //init values, and stop
       scaleHor=1.0;
       scaleVer = 1.0;
       transHor=0.0;
       transVer=0.0;
       //set zoomfactor back to 1
       zoomFactor = 1.0 ;
       return ;
     }
   //if domain does not have entries yet, also initialize and stop
   if (drawViewOwner->getDomain()->nrOfBorders() < 1)
     { //init values, and stop
       scaleHor=1.0;
       scaleVer = 1.0;
       transHor=0.0;
       transVer=0.0;
       //set zoomfactor back to 1
       zoomFactor = 1.0 ;
       return ;
     }
   //determine min and max for the domain
   double minHor = drawViewOwner->getDomain()->minCo(0);
   double maxHor = drawViewOwner->getDomain()->maxCo(0);
   double minVer = drawViewOwner->getDomain()->minCo(1);
   double maxVer = drawViewOwner->getDomain()->maxCo(1);
   double lengthHor = maxHor - minHor;
   double lengthVer = maxVer - minVer;

   //als lenghtHor/Ver fit in 100 - Width/Height, no scaling necessary. If <100 scale to 100, if > Width/Heigt scale to these
   if (lengthHor >= 100 && lengthHor< fieldWidth -10 ) {scaleHor = 1.0 ;};
   if (lengthVer >= 100 && lengthVer< fieldHeight-10 ) {scaleVer = 1.0 ;};
   if (lengthHor < 100) {scaleHor = 250 / lengthHor ;};
   if (lengthVer < 100) {scaleVer = 250 / lengthVer ;};
   if (lengthHor >= fieldWidth-10) {scaleHor = (fieldWidth -50) / lengthHor ;};
   if (lengthVer >= fieldHeight-10) {scaleVer = (fieldHeight- 50) / lengthVer ;};
   //now we have a scaled figure to fit in the view. now we translate it. The leftbottom point of the domain must be
   //after translation in 0,0
   transHor = -1.0 * minHor ;
   transVer = -1.0 * minVer ;

}
void PlotFieldView::zoomOut()
{
   if (zoomFactor <= 1.0)
     {
       zoomFactor = zoomFactor - 0.1 ;
     } else {
       zoomFactor = zoomFactor - 1.0;
     }
   if (zoomFactor == 0.0) {zoomFactor = 0.1;};
   update();              //update will invoke a paintevent that does the actual drawing
}
void PlotFieldView::zoomIn()
{
   if (zoomFactor >= 1.0)
     {
       zoomFactor = zoomFactor +1.0 ;
     } else {
       zoomFactor = zoomFactor +0.1;
     }
   update();              //update will invoke a paintevent that does the actual drawing
}
//determine the number of contours between two values and store then in an vector of integers
void PlotFieldView::contoursBetween(const double &value1, const double &value2, double *contValues, vector<int> &ContNrVect)
{  //the 0 value of contours is the minimum. Since normally only 1 points is that minimum, we don't put a contour there
   if ((value1 < contValues[1] && value2 < contValues[1]) || (value1 > contValues[nrCont] && value2 > contValues[nrCont]))
     { // no contours
       return ;
     }
    int i,j;
   //possibly contours, determine first contour smaller than value1
   for ( i = nrCont; contValues[i] > value1 ; --i){};
   //and first  one smaller than value2
   for ( j = nrCont; contValues[ j] > value2 ; --j){};
   //if those 2 values are the same, no contour between them !
   if (i==j)
     { //perhaps there is a contour through one of the values or both values ?
       if (value1==contValues[i] || value2==contValues[ i])
         {
           ContNrVect.push_back(i);
         }
       return ;
     } else {
       //there is a contour between the two, the smallest one is lower or equal to the lowest value !
       if (i<j)
         { //lowest value is i
           if (i !=0 && value1 == contValues[ i])
             { ContNrVect.push_back(i);
             }
           //all values between i and j are contours between the two values
           for ( ; i < j ; ++i) { ContNrVect.push_back(i+1); }
         } else { //lowest value is j
           if (j !=0 && value2 == contValues[ j])
             { ContNrVect.push_back( j);
             }
           //all values between j and i are contours between the two values
           for ( ; j < i ; ++j) { ContNrVect.push_back(j+1); }
         }
     }
}
//determine returnx and returny for the value contValue between value1 and value2 (linear)
bool PlotFieldView::posCont(const double &contValue, const double &x1, const double &y1 , const double &value1,
                                const double &x2,  const double &y2,  const double &value2, double &returnx , double &returny)
{ //determine relative postion of the contvalue
  if (value1==value2)
    {
     return false ;
    }
  double relpos;
  if (value1 < value2)
    { relpos = (contValue- value1) / (value2 - value1) ;
      //from point 1 we add this relative position towards point 2
      returnx = x1 + (x2-x1) * relpos;
      returny = y1 + (y2-y1) * relpos;
    } else {
      relpos = (contValue - value2) / (value1-value2) ;
      //from point 1 we add this relative position towards point 2
      returnx = x2 + (x1-x2) * relpos;
      returny = y2 + (y1-y2) * relpos;
    }
  return true;
}
