/***************************************************************************
                          topoimp.cpp  -  description
                             -------------------
    begin                : Mon May 28 2001
    copyright            : (C) 2001 by benny
    email                : bm@cage.rug.ac.be
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "topoimp.h"
#include "vectorimp.h"

//a triangle2d is an element with 3 nodes and 3 sides
TopoTriangle2d::TopoTriangle2d(TopoNode2d* n1, TopoNode2d* n2,TopoNode2d* n3,
                                                  TopoLineEdge2d* e1,TopoLineEdge2d* e2, TopoLineEdge2d* e3) : TopoElement2d(3,3)
{
   nodes[0] = n1;
   n1->addInEle(this);
   nodes[1] = n2;
   n2->addInEle(this);
   nodes[2] = n3;
   n3->addInEle(this);
   edges[0] = e1;
   e1->addNeighbour(this);
   edges[1] = e2;
   e2->addNeighbour(this);
   edges[2] = e3;
   e3->addNeighbour(this);
}
TopoTriangle2d::~TopoTriangle2d()
{
}
#ifdef LIB_WITH_QT
void TopoTriangle2d::draw(QPainter *p)
{ //we pass the command to all the edges
  //to avoid construction of an iterator, we use the fact we know a triangle has 3 edges !
  edges[0]->draw(p);
  edges[1]->draw(p);
  edges[2]->draw(p);
  //since a triangle2d does not contain nodes in the element itself, we don't have to draw nodes
}
#endif
inline double TopoTriangle2d::surface()
{  //we inline the fast path. Calculation not in an inlined function
   if (surf == 0.0 )
     { calcSurf(); };
   return surf;

}
void TopoTriangle2d::calcSurf()
{  //surface of a triangle, can be calculated out of the 3 nodes
   surf = nodes[0]->co1() *( nodes[1]->co2() - nodes[2]->co2() )
             + nodes[1]->co1() *( nodes[2]->co2() - nodes[0]->co2() )
             + nodes[2]->co1() *( nodes[0]->co2() - nodes[1]->co2() ) ;
   if (surf > 0.0 )
     {
       surf = 0.5 * surf;
      }else {
       surf = -0.5 * surf ;
      }
}
double TopoTriangle2d::localBasisFunc(int nodeI, double coor1, double coor2)
{  // a basisfunction of a triangle2d is 1 in node I, and 0 in the other points. It is a linear function
   // first determine the nodes counterclockwise. Note that nodeI is between 0 and 2 (user of function takes care of that), so module 3 does the job
   // user must also take care that coor1 and 2 are in the element triangle2d
   int j((nodeI+1)%3),  k((nodeI+2)%3) ;
   double res = ( (nodes[ j]->co1() * nodes[k]->co2()) - (nodes[ k]->co1() * nodes[ j]->co2())
                          + coor1 * (nodes[j]->co2() - nodes[k]->co2() ) + coor2 * (nodes[k]->co1() - nodes[j]->co1() )
                      ) / (surface() * 2 ) ;    //use known value of the surface in the calculation
   //since we assume coor1 and coor 2 in the triangle, we know that if res is negative the sign must be switched
   if (res < 0.0) {return -1.0*res;}
   return res;
}
double TopoTriangle2d::localBasisFuncD1(int nodeI, double coor1, double coor2)
{  //first derivative is a constant. See localBasisFunc. So coor1 and coor2 useless
   int i(nodeI), j((nodeI+1)%3), k((nodeI+2)%3) ;
   return  (nodes[j]->co2() - nodes[k]->co2() )
                  / (  nodes[ i]->co1() *( nodes[ j]->co2() - nodes[ k]->co2() )
                       + nodes[ j]->co1() *( nodes[ k]->co2() - nodes[ i]->co2() )
                       + nodes[ k]->co1() *( nodes[ i]->co2() - nodes[ j]->co2() )
                    ) ;

}
double TopoTriangle2d::localBasisFuncD2(int nodeI, double coor1, double coor2)
{  //first derivative is a constant. See localBasisFunc
   int i(nodeI), j((nodeI+1)%3), k((nodeI+2)%3) ;
   return   (nodes[k]->co1() - nodes[j]->co1() )
                  / (  nodes[i]->co1() *( nodes[j]->co2() - nodes[k]->co2() )
                       + nodes[j]->co1() *( nodes[k]->co2() - nodes[i]->co2() )
                       + nodes[k]->co1() *( nodes[i]->co2() - nodes[j]->co2() )
                    ) ;
}

bool TopoTriangle2d::testTopoTriangle2d()
{
   TopoNode2d *n1 = new TopoNodeInner2d(0.0,0.0);
   TopoNode2d *n2 = new TopoNodeInner2d(1.0,0.0);
   TopoNode2d *n3 = new TopoNodeInner2d(0.0,1.0);
   TopoLineEdge2d *e1 = new TopoLineEdge2d(n1,n2);
   TopoLineEdge2d *e2 = new TopoLineEdge2d(n2,n3);
   TopoLineEdge2d *e3 = new TopoLineEdge2d(n3,n1);
   TopoTriangle2d *t1 = new TopoTriangle2d(n1,n2,n3,e1,e2,e3);
   cerr << t1->localBasisFunc(1,1.0,0.0) << " : should be 1\n";
   cerr << t1->localBasisFunc(1,0.0,1.0) << " : should be 0 \n";
   cerr << t1->surface() << " : surface 1/2\n";
   //local matrix is automatically init to 0
   t1->setLocMat(0,0,0.01);
   t1->setLocMat(1,0,0.02);
   t1->setLocMat(2,0,0.03);
   t1->setLocMat(2,2,0.04);
   //force is not init !
   t1->setLocForce(0,0.01);
   t1->setLocForce(1,0.00);
   t1->setLocForce(2,0.02);
   cerr << t1->locMat(1,0) << " , " << t1->locMat(1,1) << " : should be 0.02 and 0.0\n";
   cerr << t1->locForce(0) << " , " << t1->locForce(1) << " : should be 0.01 and 0.0\n";
   TopoNode2d *n4 = new TopoNodeInner2d(1.0,1.0);
   TopoLineEdge2d *e4 = new TopoLineEdge2d(n2,n4);
   TopoLineEdge2d *e5 = new TopoLineEdge2d(n4,n3);
   TopoTriangle2d *t2 = new TopoTriangle2d(n2,n4,n3,e4,e5,e2);
   t1->addNeighbour(t2);
   t2->addNeighbour(t1);
   TopoNode2d *testn = t1->neighbour(0)->edge(1)->node(0);    //should be n5 !!
   cerr << testn->co1() << " , " << testn->co2() << " : should be 1.0 , 1.0 \n" ;
   //free space again
   delete n1;
   delete n2;
   delete n3;
   delete e1;
   delete e2;
   delete e3;
   delete t1;
   return true;
}

inline TopoLineEdge2d::TopoLineEdge2d(TopoNode2d* n1, TopoNode2d* n2, bool noBoundary) : TopoEdge2d(2, noBoundary)
{  //a lineedge has two nodes. In edge already 2 nodes are reserved by edge constructor
   nodes[0] = n1;
   nodes[1] = n2;
}
TopoLineEdge2d::~TopoLineEdge2d()
{ //no destruction. The controller object (a domain) does this
}
#ifdef LIB_WITH_QT
void TopoLineEdge2d::draw(QPainter *p)
{  //function requires integers, we convert our doubles to that
   p->drawLine(nodes[0]->co1(),nodes[0]->co2(), nodes[1]->co1(), nodes[1]->co2());
}
#endif

bool TopoLineEdge2d::robinEdge()
{ //a robin edge is no inneredge, and one of the nodes is a robin node. This because robin is the weakest condition, so is only sticks to the nodes if on
  //     a robin edge
   if (!innerEdge())
     {  return  (nodes[0]->robinNode() || nodes[1]->robinNode() );
     }
   return false ;
}
bool TopoLineEdge2d::neuEdge()
{ //an neuman edge if not inner, and  at least one node is neumann node, but no robin nodes (because a robin node on a neumann boundary is
  //   by force a neumann node, since that condition is stronger) !
   if (!innerEdge())
     { return ( (nodes[0]->neuNode() || nodes[1]->neuNode()) && (!nodes[0]->robinNode() ) && (! nodes[1]->robinNode() )
                  );
     }
   return false ;
}
bool TopoLineEdge2d::dirEdge()
{  //a dirichlet edge if not inner and both nodes are dirichlet (a node on a dirichlet boundary is always also dirichlet !!)
   //      *********         GENERAL REMARK       ************
  //       *  an edge that has both nodes as boundary nodes can still be an inneredge, eg a node that connects two different boundaries .
  //       *********************************************************
   if (!innerEdge())
     { return (nodes[0]->dirNode() && nodes[1]->dirNode()) ;
     }
   return false;
}

int TopoNode2d::inElement(TopoElement2d* el)
{  int nr = inEleNr(), i(0) ;
   for ( ; (i < nr) && (partOfEle[ i] != el) ; ++i ){};
   if (i==nr)
     { return -1;
     } else {
       //we found our node is in element i, what is the local node of i for that element ?
       int j(0);
       for ( ; partOfEle[i]->node(j) != this ; ++j){} ;
      return j;
     }
}

#ifdef LIB_WITH_QT
void TopoNodeInner2d::draw(QPainter *p)
{ //set appropriate color first?
  p->drawPoint(coord1,coord2);
}
void TopoNodeNeu2d::draw(QPainter *p)
{ //set appropriate color first?
  p->drawPoint(coord1,coord2);
}
void TopoNodeDir2d::draw(QPainter *p)
{ //set appropriate color first?
  p->drawPoint(coord1,coord2);
}
void TopoNodeRobin2d::draw(QPainter *p)
{ //set appropriate color first?
  p->drawPoint(coord1,coord2);
}
#endif
