// -*-c++-*-; Emacs major mode definition
/* extra changes on initial files
     - Marian Slodica : special quick classes for small matrixec or vectors, eg Matrix33
     - Benny Malengier - 05/2001 : namechange to avoid problems with STL
*/
/********************************************************************************
                R C S  --   I N F O R M A T I O N

  $Author: benny $
  $Date: 2001/06/08 09:40:42 $
  $Log: vectorimp.cpp,v $
  Revision 1.1  2001/06/08 09:40:42  benny
  *** empty log message ***

Revision 2.15  94/08/24  18:23:23  hiptmair
Automatic submission: use promoinfo for learning about changes

Revision 2.14  94/08/08  15:17:31  hiptmair
Fixed a bug in Givens rotation

Revision 2.13  94/07/21  09:24:37  hiptmair
Slight changes to prevent warning during compilation

Revision 2.12  94/04/29  23:46:16  hiptmair
Automatic submission: use promoinfo for learning about changes

Revision 2.11  94/03/14  17:31:14  hiptmair
Automatic submission: use promoinfo for learning about changes

Revision 2.10  94/03/06  22:02:24  hiptmair
Automatic submission: use promoinfo for learning about changes

Revision 2.9  94/03/03  20:22:00  hiptmair
Automatic submission: use promoinfo for learning about changes

Revision 2.8  94/03/03  11:46:22  hiptmair
Corrected mistake in vector3D::norm()

Revision 2.7  94/03/02  15:38:53  hiptmair
Large scale implementation

Revision 2.6  94/02/26  20:40:37  hiptmair
Automatic submission: use promoinfo for learning about changes

Revision 2.5  94/02/23  13:09:35  schmied
$1

Revision 2.4  94/02/22  22:34:04  hiptmair
Implementation of new methods for fullmatrix

Revision 2.3  94/02/19  18:04:59  hiptmair
Automatic submission: use promoinfo for learning about changes

Revision 2.2  94/01/05  22:00:45  hiptmair
Implemented new class FullMatrix

Revision 2.1  93/12/21  11:16:49  hiptmair
A minor syntax error has been fixed

Revision 2.0  93/12/19  18:09:34  hiptmair
Set up for version 2.0

 ********************************************************************************/
/********************************************************************************
 *                                                                              *
 *                           vector.c++                                         *
 *                                                                              *
 * Revision history:                                                            *
 *   1993    : Created by I.E. as part of the n2 mesh generator                 *
 *   05.04.93: Set up as a separate file for the finite element programme by RH *
 *   27.05.93: enhanced for three Dimensions by Th. Li.                         *
 *                                                                              *
 * Comment:                                                                     *
 *  implementation of a few remaining vector features                           *
 *                                                                              *
 ********************************************************************************/
/*===============================================================================*/

// *******************************************************************************
// GENERAL INCLUDES 
// *******************************************************************************

#include <stdlib.h>
#include <fstream.h>
#include <iomanip.h>
#include <math.h>
#include <assert.h>

// *******************************************************************************
// RELATED HEADER
// *******************************************************************************
#include <malloc.h>
#include "vectorimp.h"

// *******************************************************************************
// MEMBER FUNCTIONS: Class Vector2D
// *******************************************************************************

void Vector2D::normalize()
{
    double n = norm();
    x[0] /= n;
    x[1] /= n;
}

// *******************************************************************************
// MEMBER FUNCTIONS: Class Vector3D
// *******************************************************************************

Vector3D Vector3D::operator ^ (const Vector3D &b) const // Kreuz-Produkt
{
    return  Vector3D (  x[1]*b.x[2] - x[2]*b.x[1] ,
			x[2]*b.x[0] - x[0]*b.x[2] ,
			x[0]*b.x[1] - x[1]*b.x[0] ) ;
}


void Vector3D::normalize()
{
    double n = norm();
    x[0] /= n;
    x[1] /= n;
    x[2] /= n;
}

// *******************************************************************************
// MEMBER FUNCTIONS: Class VectornD
// *******************************************************************************

// Constructors

VectornD::VectornD(MeshCard n0,double initval) : n(n0)
{
  v = new double[n] ;
  if (v)
    {
      register int i=n;
      double *p=v;
      
      while(i--) *p++ = initval;
    }
  else
    { cerr << "Allocation failed for vector of size " << n << endl; exit(255); }
}

VectornD::VectornD(const VectornD &x) : n(x.n)
{
  v = new double[n] ;
  if (v)
    {
      register int i=n;
      double *p=v,*q=x.v;
      
      while(i--) *p++ = *q++;
    }
}

VectornD::VectornD(const double *data,unsigned int size):n(size)
{
  v = new double[n] ;
  if (v)
    {
      register int i=n;
      double *p=v;
      while(i--) *p++ = *data++;
    }
  else
    { cerr << "Allocation failed for vector of size " << n << endl; exit(255); }
}

VectornD::VectornD(const float *data,unsigned int size):n(size),v(new double [size])
{
  if (v)
    {
      register int i=n;
      double *p=v;
      while(i--) *p++ = *data++;
    }
  else
    { cerr << "Allocation failed for vector of size " << n << endl; exit(255); }
}

// destructor

VectornD::~VectornD()
{
  delete [] v;
  v = NIL;
}

// assignment operator

VectornD &VectornD::operator = (const VectornD &x)
{
#ifdef CHECK
  if (n != x.n )
    {
      cerr << "Size mismatch " << n << " <-> " << x.n << " in vector assignment!\n";
      exit(255);
    }
#else
  if (n!=x.n)
    {
      delete [] v;
      n = x.n;
      v = new double [n];
    }
#endif
  if (v)
    {
      register int i=n;
      double *p=v,*q=x.v;
      
      while(i--) *p++ = *q++;
    }
  return *this;
}

// comparison

MeshBool VectornD::operator == (const VectornD &x) const
{
#if defined(CHECK)|| defined(MV_COMPARE_DIFFERENT_SIZES)
  if (n != x.n )
    {
#ifdef COMPARE_DIFFERENT_SIZES
      return MeshFalse;
#else
      cerr << "Size mismatch " << n << " <-> " << x.n << " in vector comparison!\n";
      exit(255);
#endif
    }
  register int i = n;
#else
  register int i = n<x.n?n:x.n;
#endif
  register double *p=v,*q=x.v;
  while(i--) if (*p++ != *q++) return MeshFalse;
  return MeshTrue;
} // end operator == 

MeshBool VectornD::operator != (const VectornD &x) const
{
  return !(*this == x);
}

// Access operators

double &VectornD::operator [](int i)
{
#ifdef CHECK
  if ((i >= n) || (i<0) )
    {
      cerr << "Illegal index " << i << " for vector of size " << n << " !\n";
      exit(255);
    }
#endif
  return v[i];
}

double VectornD::operator [](int i) const
{
#ifdef CHECK
  if ((i >= n) || (i<0) )
    {
      cerr << "Illegal index " << i << " for vector of size " << n << " !\n";
      exit(255);
    }
#endif
  return v[i];
}

// Ordinary arithmetic operations that involve the creation of a new vector

VectornD VectornD::operator + (const VectornD &x) const
{
  VectornD res(x);
  
#ifdef CHECK  
  if (n != x.n)
    {
      cerr << "Vectors of different dimension cannot be added!\n";
      exit(255);
    }
  else
#endif
    {
      register int i = n;
      register double *p=res.v,*q=v;
      while(i--) *p++ += *q++;
    }
  return res;
}

VectornD VectornD::operator - (const VectornD &x) const
{
  VectornD res(*this);
  
#ifdef CHECK  
  if (n != x.n)
    {
      cerr << "Vectors of different dimension cannot be subtracted!\n";
      exit(255);
    }
  else
#endif
    {
      register int i = n;
      register double *p=res.v,*q=x.v;
      while(i--) *p++ -= *q++;
    }
  return res;
}

VectornD VectornD::operator * (double alpha) const
{
  VectornD res(*this);
  
  register int i = n;
  register double *p=res.v;
  while(i--) *p++ *= alpha;
  
  return res;
}

VectornD VectornD::operator / (double alpha) const
{
  VectornD res(*this);
  
  register int i = n;
  register double *p=res.v;
  while(i--) *p++ /= alpha;
  
  return res;
}

// friend function: multiplikation by a scalar from the left

VectornD operator * (double alpha,const VectornD &x)
{
  VectornD res(x);
  
  register int i = x.n;
  register double *p=res.v;
  while(i--) *p++ *= alpha;
  
  return res;
}

// In situ arithmetic operations

VectornD &VectornD::operator += (const VectornD &x)
{
#ifdef CHECK
  if (n != x.n )
    {
      cerr << "Size mismatch " << n << " <-> " << x.n << " in vector in situ addition!\n";
      exit(255);
    }
  register int i = n;
#else
  register int i = n<x.n ? n:x.n;
#endif
  
  register double *p = v, *q = x.v;
  while(i--) *p++ += *q++;
  
  return *this;
}

VectornD &VectornD::operator -= (const VectornD &x) 
{
#ifdef CHECK
  if (n != x.n )
    {
      cerr << "Size mismatch " << n << " <-> " << x.n << " in vector in situ subtraction!\n";
      exit(255);
    }
  register int i = n;
#else
  register int i = n<x.n ? n:x.n;
#endif
  
  register double *p = v, *q = x.v;
  while(i--) *p++ -= *q++;
  
  return *this;
}

VectornD &VectornD::operator *= (double alpha)
{
  register int i = n;
  register double *p = v;
  while(i--) *p++ *= alpha;
  return *this;
}

VectornD &VectornD::operator /= (double alpha)
{
  register int i = n;
  register double *p = v;
  while(i--) *p++ /= alpha;
  return *this;
}

// Euclidean scalar product

double VectornD::operator * (const VectornD &x) const
{
#ifdef CHECK
  if (n != x.n )
    {
      cerr << "Size mismatch " << n << " <-> " << x.n << " in scalar product!\n";
      exit(255);
    }
  register int i = n;
#else
  register int i = n<x.n ? n:x.n;
#endif
  register double s = 0;
  register double *p = v, *q = x.v;
  
  while(i--) s += *p++ * *q++;
  return s;
}

// A few combined operations that are commonplace in linear algebra routines

// x plus general alpha times y 
VectornD &VectornD::xpgay(double alpha, const VectornD &x)
{
#ifdef CHECK
  if (n != x.n )
    {
      cerr << "Size mismatch " << n << " <-> " << x.n << " in xpgay!\n";
      exit(255);
    }
  register int i = n;
#else
  register int i = n<x.n ? n:x.n;
#endif
  register double *p = v, *q = x.v;
  
  while(i--) *p++ += *q++ * alpha;
  return *this;
}

VectornD &VectornD::xmgay(double alpha, const VectornD &x)
{
#ifdef CHECK
  if (n != x.n )
    {
      cerr << "Size mismatch " << n << " <-> " << x.n << " in xmgay!\n";
      exit(255);
    }
  register int i = n;
#else
  register int i = n<x.n ? n:x.n;
#endif
  register double *p = v, *q = x.v;
  
  while(i--) *p++ -= *q++ * alpha;
  return *this;
}


VectornD &VectornD::gaxpy(double alpha, const VectornD &x)
{
#ifdef CHECK
  if (n != x.n )
    {
      cerr << "Size mismatch " << n << " <-> " << x.n << " in xpgay!\n";
      exit(255);
    }
  register int i = n;
#else
  register int i = n<x.n ? n:x.n;
#endif
  register double *p = v, *q = x.v;
  
  while(i--) *p *= alpha; *p++ += *q++;
  return *this;
}

VectornD &VectornD::neg()
{
  register int i = n;
  register double *p = v;
  while(i--) *p++ = -*p;
  return *this;
}

// Euclidean norm and its square

double VectornD::norm(void) const
{
  register int i=n;
  register double *p=v;
  register double res=0;
  
  while(i--) res += *p*(*p++);
  
  return sqrt(res);
}

double VectornD::normsq(void) const
{
  register int i=n;
  register double *p=v;
  register double res=0;
  
  while(i--) res += *p*(*p++);
  
  return res;
}

// Calculates the cosine of the angle between two vectors

double angle(const VectornD &v1,const VectornD &v2)
{
  double n1=v1.norm(),n2=v2.norm();
  
  if ((n1 > 0.0) && (n2 > 0.0))
    return ((v1*v2)/(n1*n2));
  else
    cerr << "Undefined angle for zero vector" << endl;
  return 0.0;
}

// Input - output procedures

ostream &operator << (ostream &o, const VectornD &x)
{
  o << x.n << " [ ";
  for (register int i=0; i<x.n-1; i++)
    o << x.v[i] << ' ';
  o << x.v[x.n-1] << " ]" << endl;
  return o;
}

istream &operator >> (istream &i, VectornD &x)
{
  int n0;
  char c;
  
  i >> n0 >> c;
  if (c=='[' && n0>0)
    {
      double *v = new double[n0];
      if (v)
	{
	  for (register int j=0; j<n0 && i; j++)
	    i >> v[j];
	  if (i) 
	    {
	      i >> c;
	      if (c==']') 
		{
		  delete [] x.v;
		  x.n = n0;
		  x.v = v;
		}
	      else i.clear(ios::badbit|i.rdstate());
	    }
	}
      else i.clear(ios::badbit|i.rdstate());
    }
  else i.clear(ios::badbit|i.rdstate());
  return i;
}

// *******************************************************************************
// MEMBER FUNCTIONS: Class FullMatrix
// *******************************************************************************

double FullMatrix::output_precision = 1.0E-7;

FullMatrix::FullMatrix(MeshCard i,MeshCard j,const double initval)
:i(i),j(j),size(i*(long)j)
{
  long k  ;
  a = new double[size] ;
  for(k=0;k<size;k++) 
   a[k] = initval;
}
  
FullMatrix::FullMatrix(MeshCard i,MeshCard j,const VectornD *columns)
:i(i),j(j),size(i*(long)j)
{
  register MeshCard k,l;
  a = new double[size] ;
  for(k=0;k<j;k++)
    {
#ifdef CHECK
      if (columns[k].dim() != i)
	{
	  cerr << "Illegal attempt to initialize a matrix with column size ";
	  cerr << i << " with a vector of length " << columns[k].dim() << endl;
	  assert("FullMatrix::FullMatrix ");
	}
#endif
      for(l=0;l<i;l++)
	(*this)(l,k) = (columns[k])[l];
    }
}

FullMatrix::FullMatrix(const FullMatrix &m):i(m.i),j(m.j),size(m.size)
{
  a = new double[size] ;
  for(register long l=0;l<size;l++) a[l] = m.a[l];
}

FullMatrix::~FullMatrix(void) 
{
  delete [] a;
}

FullMatrix &FullMatrix::operator =  (const FullMatrix &m)
{
#ifdef CHECK
  if ((i!=m.i) || (j!=m.j))
    {
      cerr << "No assignment of matrices of different size: ";
      cerr << i << " vs ." << m.i << ", " << j << " vs. " << m.j << endl;
      assert("&FullMatrix::operator");
    }
#endif
  for(register long l=0;l<size;l++) a[l] = m.a[l];
  return *this;
}
  
MeshBool FullMatrix::operator == (const FullMatrix &m) const
{
#ifdef CHECK
  if ((i!=m.i) || (j!=m.j))
    {
      cerr << "No comparison of matrices of different size: ";
      cerr << i << " vs ." << m.i << ", " << j << " vs. " << m.j << endl;
      assert(" FullMatrix::operator == ");
    }
#endif
  for(register long l=0;l<size;l++) if (a[l] != m.a[l]) return MeshFalse;
  return MeshTrue;
}

MeshBool FullMatrix::operator != (const FullMatrix &m) const
{
#ifdef CHECK
  if ((i!=m.i) || (j!=m.j))
    {
      cerr << "No comparison of matrices of different size: ";
      cerr << i << " vs ." << m.i << ", " << j << " vs. " << m.j << endl;
      assert("FullMatrix::operator !=  ");
    }
#endif
  for(register long l=0;l<size;l++) if (a[l] == m.a[l]) return MeshFalse;
  return MeshTrue;
}

FullMatrix::operator MeshBool ()
{
  for(register long l=0;l<size;l++) if (a[l] != 0.0) return MeshTrue;
  return MeshFalse;
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  MEMBER FUNCTION: Transposing operator ~
  +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
FullMatrix FullMatrix::operator  ~  (void) const
{
  FullMatrix res(cols(),lines());

  double *target = res.a;

  for(int l=0;l<res.j;l++) 
      { 
	double *source = (double *)(a) + l;
	int k=res.j; 
	while(k--) { *target++ = *source; source += i; }
      }
  return res;
}

FullMatrix FullMatrix::operator  +  (const FullMatrix &v) const
{
#ifdef CHECK
  if ((i!=v.i) || (j != v.j))
    {
      cerr << "Matrix size mismatch for +: ";
      cerr << i << " vs ." << v.i << ", " << j << " vs. " << v.j << endl;
      assert(" FullMatrix::operator  + ");
    }
#endif
  FullMatrix res(i,j);
  for(register long l=0;l<size;l++) res.a[l] = a[l]+v.a[l]; 
  return res;
}
FullMatrix FullMatrix::operator  -  (const FullMatrix &v) const
{
#ifdef CHECK
  if ((i!=v.i) || (j != v.j))
    {
      cerr << "Matrix size mismatch for -: ";
      cerr << i << " vs ." << v.i << ", " << j << " vs. " << v.j << endl;
      assert("FullMatrix::operator  -  ");
    }
#endif
  FullMatrix res(i,j);
  for(register long l=0;l<size;l++) res.a[l] = a[l]-v.a[l]; 
  return res;
}
FullMatrix FullMatrix::operator  *  (const FullMatrix &v) const
{
#ifdef CHECK
  if (j!=v.i)
    {
      cerr << "Matrix size mismatch for matrix multiplcation: ";
      cerr << i << " vs ." << v.i << ", " << j << " vs. " << v.j << endl;
      assert("FullMatrix::operator  *  ");
    }
#endif
  register int k,l,t;
  
  FullMatrix res(i,v.j);

  for(k=0;k<i;k++)
    for(l=0;l<v.j;l++)
      {
	double tmp = 0.0;
	for(t=0;t<j;t++) tmp += (*this)(k,t)*v(t,l);
	res(k,l) = tmp;
      }
  return res;
}

FullMatrix FullMatrix::operator  *  (double alpha) const
{
  FullMatrix res(i,j);
  for(register long l=0;l<size;l++) res.a[l] = a[l]*alpha;
  return res;
}

FullMatrix FullMatrix::operator  /  (double alpha) const
{
  FullMatrix res(i,j);
  for(register long l=0;l<size;l++) res.a[l] = a[l]/alpha;
  return res;
}

FullMatrix &FullMatrix::operator *= (double alpha)
{
  for(register long l=0;l<size;l++) a[l] *= alpha;
  return *this;
}

FullMatrix  &FullMatrix::operator /= (double alpha)
{
  for(register long l=0;l<size;l++) a[l] /= alpha;
  return *this;
}

FullMatrix  &FullMatrix::operator += (const FullMatrix &v)
{
#ifdef CHECK
  if ((i!=v.i) || (j != v.j))
    {
      cerr << "Matrix size mismatch for +=: ";
      cerr << i << " vs ." << v.i << ", " << j << " vs. " << v.j << endl;
      assert("FullMatrix::operator += ");
    }
#endif
  for(register long l=0;l<size;l++) a[l] += v.a[l]; 
  return *this;
}

FullMatrix  &FullMatrix::operator -= (const FullMatrix &v)
{
#ifdef CHECK
  if ((i!=v.i) || (j != v.j))
    {
      cerr << "Matrix size mismatch for -=: ";
      cerr << i << " vs ." << v.i << ", " << j << " vs. " << v.j << endl;
      assert("FullMatrix::oper= ");
    }
#endif
  for(register long l=0;l<size;l++) a[l] -= v.a[l]; 
  return *this;
}
VectornD FullMatrix::operator * (const VectornD &v) const
{
  VectornD res(i);
  register MeshCard k,l;
  
  const double *tvp,*vp = (double *)v;
  double *trp,*rp = (double *)res;
  const double *tmp = a;
  for(k=0;k<j;k++,tmp++)
    for(tvp=vp,trp=rp,l=0;l<i;l++)
      *(rp++) += *(vp++)*(*tmp++);
  return res;
}

#ifdef CHECK
double FullMatrix::operator()(MeshCard pi, MeshCard pj) const
{
  if ((pi >= i) || (pj >= j))
    {
      cerr << " FullMatrix::operator(): out of range: ";
      cerr << pi << " >= " << i << ", " << pj << " >= " << j << endl;
      assert("FullMatrix::operator() ");
    }
  return a[pj*i+pi];
}

double &FullMatrix::operator()(MeshCard pi, MeshCard pj)
{
  if ((pi >= i) || (pj >= j))
    {
      cerr << " FullMatrix::operator(): out of range: ";
      cerr << pi << " >= " << i << ", " << pj << " >= " << j << endl;
      assert("FullMatrix::operator() ");
    }
  return a[pj*i+pi];
}

double *FullMatrix::operator[](MeshCard pj)
{
  if (pj >= j)
    {
      cerr << " FullMatrix::operator[]: out of range: ";
      cerr << pj << " >= " << j << endl;
      assert("FullMatrix::operator[] ");
    }
  return (a+pj*i);
}
#endif

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  MEMBER FUNCTION: submat
  cuts out a submatrix with top left corner at position (pi,pj) and size
  di x dj.
  +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
FullMatrix FullMatrix::submat(MeshCard pi,MeshCard pj,MeshCard di,MeshCard dj)
{
#ifdef CHECK
  if ((pi+di > i) || (pj+dj > j))
    {
      cerr << "FullMatrix::submat: submatrix at position (" << pi << ',' << pj;
      cerr << ") of size " << di << 'x' << dj << " does not fit into ";
      cerr << "source matrix!" << endl;
      assert("FullMatrix::submat");
    }
#endif
  FullMatrix res(di,dj);
  double *target = res.a; 
  double *source = a + (pj*i+pi);
  MeshCard jump = i - di;

  while(dj--) { int k = di; while(k--) *target++ = *source++; source += jump; }
  return res;
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  MEMBER FUNCTION: lxpgaly
  line[ln_x] += (alpha * line[ln_y])
  +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
void FullMatrix::lxpgaly(double alpha,MeshCard ln_x,MeshCard ln_y)
{
#ifdef CHECK
  if ((ln_x >= i) || (ln_y >= i))
    {
      cerr << "FullMatrix::lxpgaly: indices(" << ln_x << "," << ln_y << ")";
      cerr << " illegal: " << i << " lines" << endl;
      assert("FullMatrix::lxpgaly");
    }
#endif
  register double *xl = a+ln_x;
  register double *yl = a+ln_y;
  register int jj = j; while(jj--) { *xl += (alpha * (*yl)); xl += i; yl+= i; }
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  MEMBER FUNCTION: cxpgacy
  col[cl_x] += (alpha * col[cl_y])
  +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
void FullMatrix::cxpgacy(double alpha,MeshCard cl_x,MeshCard cl_y)
{
#ifdef CHECK
  if ((cl_x >= j) || (cl_y >= j))
    {
      cerr << "FullMatrix::lxpgaly: indices(" << cl_x << "," << cl_y << ")";
      cerr << " illegal: " << j << " columns" << endl;
      assert("FullMatrix::lxpgaly");
    }
#endif
  register double *xl = a+i*cl_x;
  register double *yl = a+i*cl_y;
  register int jj = i; while(jj--) { *xl += (alpha * (*yl)); xl++; yl++; }
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  MEMBER FUNCTION: rotl
  line[ln_x] = line[ln_x] * c - line[ln_y]*s
  line[ln_y] = line[ln_x] * s + line[ln_y]*c
  rotation of a pair of matrix rows
  +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
void FullMatrix::rotl(double c,double s,MeshCard ln_x,MeshCard ln_y)
{
#ifdef CHECK
  if ((ln_x >= i) || (ln_y >= i))
    {
      cerr << "FullMatrix::rotl: indices(" << ln_x << "," << ln_y << ")";
      cerr << " illegal: " << i << " lines" << endl;
      assert("FullMatrix::rotl");
    }
#endif
  register double tmp;
  register double *xl = a+ln_x;
  register double *yl = a+ln_y;
  register int jj = j; 
  while(jj--) 
    { 
      tmp = *xl;
      *xl = c*tmp-s*(*yl);
      *yl = s*tmp+c*(*yl);
      xl += i; yl+= i; 
    }
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  MEMBER FUNCTION: rotc
  col[cl_x] = col[cl_x] * c - col[cl_y]*s
  col[cl_y] = col[cl_x] * s + col[cl_y]*c
  rotation of a pair of matrix rows
  +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
void FullMatrix::rotc(double c,double s,MeshCard cl_x,MeshCard cl_y)
{
#ifdef CHECK
  if ((cl_x >= j) || (cl_y >= j))
    {
      cerr << "FullMatrix::rotl: indices(" << cl_x << "," << cl_y << ")";
      cerr << " illegal: " << j << " columns" << endl;
      assert("FullMatrix::rotc");
    }
#endif
  register double tmp;
  register double *xl = a+i*cl_x;
  register double *yl = a+i*cl_y;
  register int jj = i; 
  while(jj--) 
    { 
      tmp = *xl;
      *xl = c*tmp-s*(*yl);
      *yl = s*tmp+c*(*yl);
      xl++; yl++; 
    }
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  MEMBER FUNCTION: multl
  line[ln] *= alpha
  +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
void FullMatrix::multl(double alpha,MeshCard ln)
{
#ifdef CHECK
  if (ln >= i)
    {
      cerr << "FullMatrix::multl: index(" << ln << ")";
      cerr << " illegal: " << i << " lines" << endl;
      assert("FullMatrix::multl");
    }
#endif
  register double *xl = a+ln;
  register int jj = j; while(jj--) { *xl *= alpha; xl += i; }
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  MEMBER FUNCTION: multc
  col[col] *= alpha
  +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
void FullMatrix::multc(double alpha,MeshCard col)
{
#ifdef CHECK
  if (col >= j)
    {
      cerr << "FullMatrix::multc: index(" << col << ")";
      cerr << " illegal: " << j << " columns" << endl;
      assert("FullMatrix::multc");
    }
#endif
  register double *xl = a+i*col;
  register int jj = i; while(jj--) { *xl *= alpha; xl++; }
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  MEMBER FUNCTION: swapl
  line[ln_1] <-> line[ln_2]
  +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
void FullMatrix::swapl(MeshCard ln_1,MeshCard ln_2)
{
#ifdef CHECK
  if ((ln_1 >= i) || (ln_2 >= i))
    {
      cerr << "FullMatrix::lxpgaly: indices(" << ln_1 << "," << ln_2 << ")";
      cerr << " illegal: " << i << " lines" << endl;
      assert("FullMatrix::swapl");
    }
#endif
  register double tmp;
  register double *xl = a+ln_1;
  register double *yl = a+ln_2;
  register int jj = j; while(jj--) { tmp = *xl; *xl = *yl; *yl = tmp; xl += i; yl+= i; }
}

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  MEMBER FUNCTION: swapc
  col[cl_1] <-> col[cl_2]
  +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
void FullMatrix::swapc(MeshCard cl_1,MeshCard cl_2)
{
#ifdef CHECK
  if ((cl_1 >= j) || (cl_2 >= j))
    {
      cerr << "FullMatrix::lxpgaly: indices(" << cl_1 << "," << cl_2 << ")";
      cerr << " illegal: " << j << " columns" << endl;
      assert("FullMatrix::swapc");
    }
#endif
  register double tmp;
  register double *xl = a+i*cl_1;
  register double *yl = a+i*cl_2;
  register int jj = i; while(jj--) { tmp = *xl; *xl = *yl; *yl = tmp; xl++; yl++; }
}

// <<<<<<<<<<<< FRIEND FUNCTIONS >>>>>>>>>>>>
FullMatrix operator  *  (double alpha,const FullMatrix &m)
{
  FullMatrix res(m.i,m.j);
  for(register long l=0;l<m.size;l++) res.a[l] = alpha*m.a[l];
  return res;
}
ostream &operator << (ostream &o, const FullMatrix &m)
{
  int k,l;
  for(k=0;k<m.i;k++)
    {
      o << "[ ";
      for(l=0;l<m.j;l++) 
	{
	  o << setw(9);
	  o << setprecision(4);
	  o << ((fabs(m(k,l)) > FullMatrix::output_precision)?(m(k,l)):0) << " ";
	  if (m(k,l) >= 0) o << ' ';
	}
      o << "]" << endl;
    }
  return o;
}

istream &operator >> (istream &i, FullMatrix &)
{
  cerr << "istream &operator >> (istream &i, FullMatrix &x):";
  cerr << " sorry not implemented" << endl;
  return i;
}

// *******************************************************************************
// MEMBER FUNCTIONS: Class Matrix22
// *******************************************************************************

double Matrix22::output_precision = 1.0E-7;

Matrix22::Matrix22(double initval)
{
  a = new double[4] ;
  int k  ;
  for(k = 0; k<2; k++)
    a[k] = 0.0 ;
}

  Matrix22::Matrix22(const Vector2D *columns)
  {
  a = new double[4] ;
  register MeshCard k,l;
  for(k=0;k<2;k++)
    {
      for(l=0;l<2;l++)
        (*this)(l,k) = (columns[k])[l];
    }
  }

Matrix22::Matrix22(const Matrix22 &m)
{
  a = new double[4] ;
  for(register long l=0;l<4;l++) a[l] = m.a[l];
}

Matrix22::~Matrix22(void)
{
  delete [] a;
  a=NIL;
;
}

Matrix22 &Matrix22::operator =  (const Matrix22 &m)
{
  for(register long l=0;l<4;l++) a[l] = m.a[l];
  return *this;
}
 
MeshBool Matrix22::operator == (const Matrix22 &m) const
{
  for(register long l=0;l<4;l++) if (a[l] != m.a[l]) return MeshFalse;
  return MeshTrue;
}
 
MeshBool Matrix22::operator != (const Matrix22 &m) const
{
  for(register long l=0;l<4;l++) if (a[l] == m.a[l]) return MeshFalse;
  return MeshTrue;
}
 
Matrix22::operator MeshBool ()
{
  for(register long l=0;l<4;l++) if (a[l] != 0.0) return MeshTrue;
  return MeshFalse;
}


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  MEMBER FUNCTION: Transposing operator ~
  +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
Matrix22 Matrix22::operator  ~  (void) const
{
  Matrix22 res(0.0);

  double *target = res.a;

  for(int l=0;l<2;l++)
      {  
        double *source = (double *)(a) + l;

        int k=2;
        while(k--) { *target++ = *source; source += 2; }
      }  
  return res;
}

Matrix22 Matrix22::operator  +  (const Matrix22 &v) const
{
  Matrix22 res(0.0);
  for(register long l=0;l<4;l++) res.a[l] = a[l]+v.a[l];
  return res;
}   

Matrix22 Matrix22::operator  -  (const Matrix22 &v) const
{
  Matrix22 res(0.0);
  for(register long l=0;l<4 ;l++) res.a[l] = a[l]-v.a[l];
  return res;
}  

Matrix22 Matrix22::operator  *  (const Matrix22 &v) const
{
  register int k,l,t;

  Matrix22 res(0.0);
 
  for(k=0;k<2;k++)
    for(l=0;l<2;l++)
      {
        double tmp = 0.0;
        for(t=0;t<2;t++) tmp += (*this)(k,t)*v(t,l);
        res(k,l) = tmp;
      }
  return res;
}
 
Matrix22 Matrix22::operator  *  (double alpha) const
{
  Matrix22 res(0.0);
  for(register long l=0;l<4;l++) res.a[l] = a[l]*alpha;
  return res;
}
 
Matrix22 Matrix22::operator  /  (double alpha) const
{
  Matrix22 res(0.0);
  for(register long l=0;l<4;l++) res.a[l] = a[l]/alpha;
  return res;
}
 
Matrix22 &Matrix22::operator *= (double alpha)
{
  for(register long l=0;l<4;l++) a[l] *= alpha;
  return *this;
}
 
Matrix22  &Matrix22::operator /= (double alpha)
{
  for(register long l=0;l<4;l++) a[l] /= alpha;
  return *this;
}

Matrix22  &Matrix22::operator += (const Matrix22 &v)
{
  for(register long l=0;l<4;l++) a[l] += v.a[l];
  return *this;
}
 
Matrix22  &Matrix22::operator -= (const Matrix22 &v)
{
  for(register long l=0;l<4;l++) a[l] -= v.a[l];
  return *this;
}

double Matrix22::determinant(void)
{   return  a[0]*a[3] - a[1]*a[2] ; }

void Matrix22::repcol(MeshCard col, Vector2D b)
//replace i-th collumn by Vector2D b
{
   uint i;
   for(i=0;i<2;i++)
      a[2*col+i] = b[i];
}
 
 
#ifdef CHECK

double Matrix22::operator()(MeshCard pi, MeshCard pj) const
{
  if ((pi >= 2) || (pj >= 2))
    {
      cerr << " Matrix22::operator(): out of range: ";
      cerr << pi << " >= " << 2 << ", " << pj << " >= " << 2 << endl;
      assert("Matrix22::operator() ");
    }
  return a[pj*2+pi];
}

double &Matrix22::operator()(MeshCard pi, MeshCard pj)
{
  if ((pi >= 2) || (pj >= 2))
    {
      cerr << " Matrix22::operator(): out of range: ";
      cerr << pi << " >= " << 2 << ", " << pj << " >= " << 2 << endl;
      assert("Matrix22::operator() ");
    }
  return a[pj*2+pi];
}   
 
double *Matrix22::operator[](MeshCard pj)
{
  if (pj >= 2)
    {
      cerr << " Matrix22::operator[]: out of range: ";
      cerr << pj << " >= " << 2 << endl;
      assert("Matrix22::operator[] ");
    }
  return (a+pj*2);
}   
#endif

// <<<<<<<<<<<< FRIEND FUNCTIONS >>>>>>>>>>>>

Matrix22 operator  *  (double alpha,const Matrix22 &m)
{
  Matrix22 res(0.0);
  for(register long l=0;l<4;l++) res.a[l] = alpha*m.a[l];
  return res;
}

ostream &operator << (ostream &o, const Matrix22 &m)
{
  int k,l;
  for(k=0;k<2;k++)
    {
      o << "[ ";
      for(l=0;l<2;l++)
        {
          o << setw(9);
          o << setprecision(4);
          o << ((fabs(m(k,l)) > Matrix22::output_precision)?(m(k,l)):0) << " ";
          if (m(k,l) >= 0) o << ' ';
        }
      o << "]" << endl;
    }
  return o;
}

istream &operator >> (istream &i, Matrix22 &)
{
  cerr << "istream &operator >> (istream &i, Matrix22 &x):";
  cerr << " sorry not implemented" << endl;
  return i;
}

Vector2D operator * (const Matrix22& a, const Vector2D& b)
//Matrix * Vector
{ uint i;
  Vector2D res  ;
  for(i=0;i<2;i++)
    res.x[i] = a(i,0)*b[uint(0)] + a(i,1)*b[uint(1)] ;
    return res ;
}



// *******************************************************************************
// MEMBER FUNCTIONS: Class Matrix33
// *******************************************************************************

double Matrix33::output_precision = 1.0E-7;

Matrix33::Matrix33(double initval)
{
a = new double[9] ;
int k  ;
for(k = 0; k<9; k++)
  a[k] = 0.0 ;
}

Matrix33::Matrix33(const Vector3D *columns)
{
a = new double[9] ;
register MeshCard k,l;
for(k=0;k<3;k++)
  {
    for(l=0;l<3;l++)
      (*this)(l,k) = (columns[k])[l];
  }
}

Matrix33::Matrix33(const Matrix33 &m)
{
a = new double[9] ;
for(register long l=0;l<9;l++) a[l] = m.a[l];
}

Matrix33::~Matrix33(void)
{
  delete [] a;
  a = NIL;
}

Matrix33 &Matrix33::operator =  (const Matrix33 &m)
{
  for(register long l=0;l<9;l++) a[l] = m.a[l];
  return *this;
}
 
MeshBool Matrix33::operator == (const Matrix33 &m) const
{
  for(register long l=0;l<9;l++) if (a[l] != m.a[l]) return MeshFalse;
  return MeshTrue;
}
 
MeshBool Matrix33::operator != (const Matrix33 &m) const
{
  for(register long l=0;l<9;l++) if (a[l] == m.a[l]) return MeshFalse;
  return MeshTrue;
}
 
Matrix33::operator MeshBool ()
{
  for(register long l=0;l<9;l++) if (a[l] != 0.0) return MeshTrue;
  return MeshFalse;
}


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  MEMBER FUNCTION: Transposing operator ~
  +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
Matrix33 Matrix33::operator  ~  (void) const
{
  Matrix33 res(0.0);

  double *target = res.a;

  for(int l=0;l<3;l++)
      {  
        double *source = (double *)(a) + l;
        int k=3;
        while(k--) { *target++ = *source; source += 3; }
      }  
  return res;
}

Matrix33 Matrix33::operator  +  (const Matrix33 &v) const
{
  Matrix33 res(0.0);
  for(register long l=0;l<9;l++) res.a[l] = a[l]+v.a[l];
  return res;
}   

Matrix33 Matrix33::operator  -  (const Matrix33 &v) const
{
  Matrix33 res(0.0);
  for(register long l=0;l<9 ;l++) res.a[l] = a[l]-v.a[l];
  return res;
}  

Matrix33 Matrix33::operator  *  (const Matrix33 &v) const
{
  register int k,l,t;

  Matrix33 res(0.0);
 
  for(k=0;k<3;k++)
    for(l=0;l<3;l++)
      {
        double tmp = 0.0;
        for(t=0;t<3;t++) tmp += (*this)(k,t)*v(t,l);
        res(k,l) = tmp;
      }
  return res;
}
 
Matrix33 Matrix33::operator  *  (double alpha) const
{
  Matrix33 res(0.0);
  for(register long l=0;l<9;l++) res.a[l] = a[l]*alpha;
  return res;
}
 
Matrix33 Matrix33::operator  /  (double alpha) const
{
  Matrix33 res(0.0);
  for(register long l=0;l<9;l++) res.a[l] = a[l]/alpha;
  return res;
}
 
Matrix33 &Matrix33::operator *= (double alpha)
{
  for(register long l=0;l<9;l++) a[l] *= alpha;
  return *this;
}
 
Matrix33  &Matrix33::operator /= (double alpha)
{
  for(register long l=0;l<9;l++) a[l] /= alpha;
  return *this;
}

Matrix33  &Matrix33::operator += (const Matrix33 &v)
{
  for(register long l=0;l<9;l++) a[l] += v.a[l];
  return *this;
}
 
Matrix33  &Matrix33::operator -= (const Matrix33 &v)
{
  for(register long l=0;l<9;l++) a[l] -= v.a[l];
  return *this;
}

double Matrix33::determinant(void)
{
   double deta;
   deta = a[0]*a[4]*a[8] +
          a[1]*a[5]*a[6] +
          a[2]*a[3]*a[7] -
          a[2]*a[4]*a[6] -
          a[5]*a[7]*a[0] -
          a[8]*a[1]*a[3];
   return deta;
}
 
void Matrix33::repcol(MeshCard col, Vector3D b)
//replace i-th collumn by Vector3D b
{
   uint i;
   for(i=0;i<3;i++)
      a[3*col+i] = b[i];
}
 
 
#ifdef CHECK

double Matrix33::operator()(MeshCard pi, MeshCard pj) const
{
  if ((pi >= 3) || (pj >= 3))
    {
      cerr << " Matrix33::operator(): out of range: ";
      cerr << pi << " >= " << 3 << ", " << pj << " >= " << 3 << endl;
      assert("Matrix33::operator() ");
    }
  return a[pj*3+pi];
}

double &Matrix33::operator()(MeshCard pi, MeshCard pj)
{
  if ((pi >= 3) || (pj >= 3))
    {
      cerr << " Matrix33::operator(): out of range: ";
      cerr << pi << " >= " << 3 << ", " << pj << " >= " << 3 << endl;
      assert("Matrix33::operator() ");
    }
  return a[pj*3+pi];
}   
 
double *Matrix33::operator[](MeshCard pj)
{
  if (pj >= 3)
    {
      cerr << " Matrix33::operator[]: out of range: ";
      cerr << pj << " >= " << 3 << endl;
      assert("Matrix33::operator[] ");
    }
  return (a+pj*3);
}   
#endif

// <<<<<<<<<<<< FRIEND FUNCTIONS >>>>>>>>>>>>

Matrix33 operator  *  (double alpha,const Matrix33 &m)
{
  Matrix33 res(0.0);
  for(register long l=0;l<9;l++) res.a[l] = alpha*m.a[l];
  return res;
}

ostream &operator << (ostream &o, const Matrix33 &m)
{
  int k,l;
  for(k=0;k<3;k++)
    {
      o << "[ ";
      for(l=0;l<3;l++)
        {
          o << setw(9);
          o << setprecision(4);
          o << ((fabs(m(k,l)) > Matrix33::output_precision)?(m(k,l)):0) << " ";
          if (m(k,l) >= 0) o << ' ';
        }
      o << "]" << endl;
    }
  return o;
}

istream &operator >> (istream &i, Matrix33 &)
{
  cerr << "istream &operator >> (istream &i, Matrix33 &x):";
  cerr << " sorry not implemented" << endl;
  return i;
}

Vector3D operator * (Matrix33& a, Vector3D& b)
//Matrix * Vector
{ uint i;
  Vector3D res  ;
  for(i=0;i<3;i++)
    res.x[i] = a(i,0)*b[uint(0)] + a(i,1)*b[uint(1)] + a(i,2)*b[uint(2)] ;
    return res ;
}




 




/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  GLOBAL FUNCTION: SolveLin
  1st arg: (FullMat &) regular N x (N+1) matrix whose last column contains the 
                      right hand side
  The global function solves square systems of linear equations by applying 
  Gaussian elimination with a simple pivot strategy. FALSE is returned if the
  systems turns  out to be singular
  +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
MeshBool SolveLin(FullMatrix &mat,double _eps)
{
  int i,j,k;    /* Index variables */
  double max;   /* Temporary variable for searching the pivotal entry */
  int dim = mat.lines(); /* size of linear system */
  
#ifdef CHECK
  if (mat.cols() != (dim+1))
    {
      cerr << "SolvLin: No square matrix supplied! :";
      cerr << "Dimensions: " << dim << " x " << mat.cols() << endl;
      assert("SolvLin");
    }
#endif
  for(i=0;i<dim;i++)
    {
      max = 0.0;
      
      /* Searching the current column for a pivot */
      
      for(k=j=i;j<dim;j++)
	if (fabs(mat(j,i))> max) { max = fabs(mat(j,i)); k=j; }
      
      if (max < _eps) return(MeshFalse);
      
      /* Swapping lines */
      if (k != i) mat.swapl(i,k);
      
      /* Elimination */
      double pivot;
      for(pivot = mat(i,i),j=i+1;j<dim;j++)
	mat.lxpgaly(-mat(j,i)/pivot,j,i);
    }

  /* Backsubstitution */
  mat[dim-1][dim] /= mat[dim-1][dim-1];
  for(i=dim-2;i>=0;i--)
    {
      for (k=i+1;k<dim;k++) mat(i,dim) -= mat(k,dim)*mat(i,k);
      
      mat(i,dim) /= mat(i,i);
    }
  return(MeshTrue);
} /* end SolveLin */

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  GLOBAL FUNCTION: Givens (first version)
  1st arg: (double *) vector with two components to be rotated.
  2nd arg: (double *) buffer for returning the crucial sin and cos entries of the
           rotation matrix
  this routine performs the Givens rotation wrenching the original vector into
  a direction parallel to the x-Axis.
  +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
MeshBool Givens(const double *vec,double *rot,double accuracy)
{
  double     r  ;  /* temporary variable */ 
  double abs1 = fabs(vec[1]);
  double abs0 = fabs(vec[0]);
  
  if (abs1 < accuracy) { rot[0] = 1.0; rot[1] = 0.0; }
  else
    {
      if(abs1 > abs0)
      {
	r = -vec[0]/vec[1];
	rot[1] = 1/sqrt(1+r*r);
	rot[0] = rot[1]*r;
      }
      else
	{
	  r = -vec[1]/vec[0];
	  rot[0] = 1/sqrt(1+r*r);
	  rot[1] = rot[0]*r;
	}
    }
  return MeshTrue;
} /* end Givens */

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  GLOBAL FUNCTION: Givens (second version)
  1st arg: (float *) vector with two components to be rotated.
  2nd arg: (double *) buffer for returning the crucial sin and cos entries of the
  rotation matrix
  this routine performs the Givens rotation wrenching the original vector into
  a direction parallel to the x-Axis.
  +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
MeshBool Givens(const float *vec,double *rot,double accuracy)
{
  double     r  ;  /* temporary variable */ 
  
  if (fabs(vec[1]) < accuracy) { rot[0] = 1.0; rot[1] = 0.0; }
  else
    {
      if(fabs(vec[1]) > fabs(vec[0]))
        {
	  r = -vec[0]/vec[1];
	  rot[1] = 1/sqrt(1+r*r);
	  rot[0] = rot[1]*r;
        }
      else
        {
	  r = -vec[1]/vec[0];
	  rot[0] = 1/sqrt(1+r*r);
	  rot[1] = rot[0]*r;
        }
    }
  return MeshTrue;
} /* end Givens */

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  GLOBAL FUNCTION: Givens (second version)
  1st arg: (double *) vector with two components to be rotated.
  2nd arg: (float *) buffer for returning the crucial sin and cos entries of the
  rotation matrix
  this routine performs the Givens rotation wrenching the original vector into
  a direction parallel to the x-Axis.
  +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
MeshBool Givens(const double *vec,float *rot,double accuracy)
{
  double     r  ;  /* temporary variable */ 
  
  if (fabs(vec[1]) < accuracy) { rot[0] = 1.0; rot[1] = 0.0; }
  else
    {
      if(fabs(vec[1]) > fabs(vec[0]))
        {
	  r = -vec[0]/vec[1];
	  rot[1] = 1/sqrt(1+r*r);
	  rot[0] = rot[1]*r;
        }
      else
        {
	  r = -vec[1]/vec[0];
	  rot[0] = 1/sqrt(1+r*r);
	  rot[1] = rot[0]*r;
        }
    }
  return MeshTrue;
} /* end Givens */

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  GLOBAL FUNCTION: Givens (second version)
  1st arg: (float *) vector with two components to be rotated.
  2nd arg: (float *) buffer for returning the crucial sin and cos entries of the
  rotation matrix
  this routine performs the Givens rotation wrenching the original vector into
  a direction parallel to the x-Axis.
  +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
MeshBool Givens(const float *vec,float *rot,double accuracy)
{
  double     r  ;  /* temporary variable */ 
  
  if (fabs(vec[1]) < accuracy) { rot[0] = 1.0; rot[1] = 0.0; }
  else
    {
      if(fabs(vec[1]) > fabs(vec[0]))
        {
	  r = -vec[0]/vec[1];
	  rot[1] = 1/sqrt(1+r*r);
	  rot[0] = rot[1]*r;
        }
      else
        {
	  r = -vec[1]/vec[0];
	  rot[0] = 1/sqrt(1+r*r);
	  rot[1] = rot[0]*r;
        }
    }
  return MeshTrue;
} /* end Givens */

//#define PEDBG
//#define XXDBG
/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  GLOBAL FUNCTION: partElim
  1st arg: (FullMatrix &) matrix on which elimination should be performed
  2nd arg: (MeshCard) terminating column for elimination
  3rd arg: (MeshCard) starting column for elimination
  4th arg: (double) threshold for dismissing a matrix entry a zero
  This method eliminates all below diagonal entries of the specified columns
  by means of Gaussian elimination without pivot search.
  +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
MeshBool partElim(FullMatrix &mat,MeshCard right,MeshCard left,double
#ifdef CHECK
		  accuracy
#endif
		  )
{
#ifdef XXDBG
  FullMatrix tmp_mat(mat);
#endif
  MeshCard i,j;
  MeshCard ldim = mat.lines();
  MeshCard cdim = mat.cols();
  double pivot;
#ifdef CHECK
  if ((right >= cdim) || (left > right))
    {
      cerr << "partElim: Illegal columns " << left << " - " << right;
      cerr << " for " << ldim << " x " << cdim << " matrix" << endl;
      assert("partElim");
    }
#endif
  for(j=left;j<=right;j++)
    {
      for(i=j+1;i<ldim;i++)
	{
	  pivot = mat(j,j);
#ifdef CHECK
	  if (fabs(pivot) < accuracy)
	    {
	      cerr << "Vanishing pivot (" << j << ',' << j;
	      cerr << ") = " << pivot << endl;
#ifdef XXDBG
	      cerr << "Original matrix: " << endl << tmp_mat << endl;
	      cerr << "Eliminated matrix: " << endl << mat << endl;
#endif
	      return MeshFalse;
	    }
#endif
	  
#ifdef PEDBG
	  cout << "Elimination based on (" << j << ',' << j << ") = ";
	  cout << mat(j,j) << " and (" << i << ',' << j << ") = ";
	  cout << mat(i,j) << endl;
#endif
	  mat.lxpgaly(-mat(i,j)/pivot,i,j);
	  
#ifdef PEDBG
	  cout << "Matrix after elimination: " << endl << mat << endl;
	  cout << "----------------------------------------" << endl;
#endif
	}
    }
  return MeshTrue;
} // end partElim   

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  GLOBAL FUNCTION: partElimQR
  1st arg: (FullMatrix &) matrix on which elimination should be performed
  2nd arg: (MeshCard) terminating column for elimination
  3rd arg: (MeshCard) starting column for elimination
  4th arg: (double) threshold for dismissing a matrix entry a zero
  This method eliminates all below diagonal entries of the specified columns
  by means of Givens rotation. 
  +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
MeshBool partElimQR(FullMatrix &mat,MeshCard right,MeshCard left,double)
{
  MeshCard i,j;
  MeshCard ldim = mat.lines();
  MeshCard cdim = mat.cols();
  double vec[2],rot[2];
#ifdef CHECK
  if ((right >= cdim) || (left > right))
    {
      cerr << "partElim: Illegal columns " << left << " - " << right;
      cerr << " for " << ldim << " x " << cdim << " matrix" << endl;
      assert("partElim");
    }
#endif
  for(j=left;j<=right;j++)
    {
      for(i=j+1;i<ldim;i++)
	{
	  vec[0] = mat(j,j); vec[1] = mat(i,j);
	  if (!Givens(vec,rot)) 
	    {
	      cerr << "Matrix singular!" << endl;
	      return MeshFalse;
	    }
#ifdef PEDBG
	  cout << "Rotation based on (" << j << ',' << j << ") = ";
	  cout << mat(j,j) << " and (" << i << ',' << j << ") = ";
	  cout << mat(i,j);
	  cout << " [c= " << rot[0] << ", s= " << rot[1] << " ]" << endl;
#endif
	  mat.rotl(rot[0],rot[1],j,i);
#ifdef PEDBG
	  cout << "Matrix after rotation: " << endl << mat << endl;
	  cout << "----------------------------------------" << endl;
#endif
	}
    }
  return MeshTrue;
} // end partElim   

// end of file
